# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Defines ToggleColumnMixIn class
"""
from __future__ import print_function

import logging

from argos.qt import QtCore, QtWidgets, Qt

logger = logging.getLogger(__name__)

# Qt classes have many ancestors
#pylint: disable=R0901

# Don't want to create constructors for this mixin just to satisfy Pylint so
# we disable W0201 (attribute-defined-outside-init)
#pylint: disable=W0201


class ToggleColumnMixIn(object):
    """ Adds actions to a QTableView that can show/hide columns
        by right clicking on the header.

        Has functionality for reading/writing from persitent settings.
    """
    def addHeaderContextMenu(self, checked = None, checkable = None, enabled = None):
        """ Adds the context menu from using header information

            checked can be a header_name -> boolean dictionary. If given, headers
            with the key name will get the checked value from the dictionary.
            The corresponding column will be hidden if checked is False.

            checkable can be a header_name -> boolean dictionary. If given, header actions
            with the key name will get the checkable value from the dictionary. (Default True)

            enabled can be a header_name -> boolean dictionary. If given, header actions
            with the key name will get the enabled value from the dictionary. (Default True)
        """
        checked = checked if checked is not None else {}
        checkable = checkable if checkable is not None else {}
        enabled = enabled if enabled is not None else {}

        horizontal_header = self.horizontalHeader()
        horizontal_header.setContextMenuPolicy(Qt.ActionsContextMenu)

        self.toggle_column_actions_group = QtWidgets.QActionGroup(self)
        self.toggle_column_actions_group.setExclusive(False)
        self.__toggle_functions = []  # for keeping references

        for col in range(horizontal_header.count()):
            column_label = self.model().headerData(col, Qt.Horizontal, Qt.DisplayRole)
            #logger.debug("Adding: col {}: {}".format(col, column_label))
            action = QtWidgets.QAction(str(column_label),
                                   self.toggle_column_actions_group,
                                   checkable = checkable.get(column_label, True),
                                   enabled = enabled.get(column_label, True),
                                   toolTip = "Shows or hides the {} column".format(column_label))
            func = self.__makeShowColumnFunction(col)
            self.__toggle_functions.append(func) # keep reference
            horizontal_header.addAction(action)
            is_checked = checked.get(column_label, not horizontal_header.isSectionHidden(col))
            horizontal_header.setSectionHidden(col, not is_checked)
            action.setChecked(is_checked)
            action.toggled.connect(func)


    def getHeaderContextMenuActions(self):
        """ Returns the actions of the context menu of the header
        """
        return self.horizontalHeader().actions()


    def __makeShowColumnFunction(self, column_idx):
        """ Creates a function that shows or hides a column."""
        show_column = lambda checked: self.setColumnHidden(column_idx, not checked)
        return show_column


    def readViewSettings(self, key, settings=None):
        """ Reads the persistent program settings

            :param key: key where the setting will be read from
            :param settings: optional QSettings object which can have a group already opened.
            :returns: True if the header state was restored, otherwise returns False
        """
        #logger.debug("Reading view settings for: {}".format(key))
        if settings is None:
            settings = QtCore.QSettings()

        horizontal_header = self.horizontalHeader()
        header_restored = horizontal_header.restoreState(settings.value(key))

        # update actions
        for col, action in enumerate(horizontal_header.actions()):
            is_checked = not horizontal_header.isSectionHidden(col)
            action.setChecked(is_checked)

        return header_restored

    def saveProfile(self, key, settings=None):
        """ Writes the view settings to the persistent store
            :param key: key where the setting will be read from
            :param settings: optional QSettings object which can have a group already opened.
        """
        #logger.debug("Writing view settings for: {}".format(key))
        if settings is None:
            settings = QtCore.QSettings()
        settings.setValue(key, self.horizontalHeader().saveState())



class ToggleColumnTableWidget(QtWidgets.QTableWidget, ToggleColumnMixIn):
    """ A QTableWidget where right clicking on the header allows the user to show/hide columns
    """
    pass



class ToggleColumnTableView(QtWidgets.QTableView, ToggleColumnMixIn):
    """ A QTableView where right clicking on the header allows the user to show/hide columns
    """
    pass



class ToggleColumnTreeWidget(QtWidgets.QTreeWidget, ToggleColumnMixIn):
    """ A QTreeWidget where right clicking on the header allows the user to show/hide columns
    """
    def horizontalHeader(self):
        """ Returns the horizontal header (of type QHeaderView).

            Override this if the horizontalHeader() function does not exist.
        """
        return self.header()



class ToggleColumnTreeView(QtWidgets.QTreeView, ToggleColumnMixIn):
    """ A QTreeView where right clicking on the header allows the user to show/hide columns
    """
    def horizontalHeader(self):
        """ Returns the horizontal header (of type QHeaderView).

            Override this if the horizontalHeader() function does not exist.
        """
        return self.header()


#pylint: enable=R0901





# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Contains classes and functions based on Qt. These could in principle be useful for
    other projects but would then require minor tweaking (e.g. imports).
"""
# Import commonly used function into the package name space for convenience.

from argos.qt.misc import Qt, QtCore, QtGui, QtWidgets, QtSvg, QtSignal, QtSlot
from argos.qt.misc import initQApplication, initQCoreApplication



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

"""
    Main window functionality

"""

from __future__ import absolute_import, division, print_function

import sys
from functools import partial

from argos.collect.collector import Collector
from argos.config.abstractcti import ctiDumps, ctiLoads
from argos.config.abstractcti import AbstractCti
from argos.config.configtreemodel import ConfigTreeModel
from argos.config.configtreeview import ConfigWidget
from argos.info import DEBUGGING, PROJECT_NAME
from argos.inspector.abstract import AbstractInspector, UpdateReason
from argos.inspector.dialog import OpenInspectorDialog
from argos.inspector.registry import InspectorRegItem
from argos.inspector.selectionpane import InspectorSelectionPane, addInspectorActionsToMenu
from argos.qt import Qt, QtCore, QtGui, QtWidgets, QtSignal, QtSlot

from argos.repo.detailplugins.attr import AttributesPane
from argos.repo.detailplugins.dim import DimensionsPane
from argos.repo.detailplugins.prop import PropertiesPane
from argos.repo.repotreeview import RepoWidget
from argos.repo.testdata import createArgosTestData
from argos.utils.cls import check_class
from argos.utils.misc import string_to_identifier
from argos.widgets.aboutdialog import AboutDialog
from argos.widgets.constants import CENTRAL_MARGIN, CENTRAL_SPACING
from argos.widgets.pluginsdialog import PluginsDialog
import logging


logger = logging.getLogger(__name__)

# The main window inherits from a Qt class, therefore it has many
# ancestors public methods and attributes.
# pylint: disable=R0901, R0902, R0904, W0201


class MainWindow(QtWidgets.QMainWindow):
    """ Main application window.
    """
    __numInstances = 0

    # Emitted when the inspector has changed.
    sigInspectorChanged = QtSignal(InspectorRegItem)

    def __init__(self, argosApplication):
        """ Constructor
            :param reset: If true the persistent settings, such as column widths, are reset.
        """
        super(MainWindow, self).__init__()
        self._windowNumber = MainWindow.__numInstances # Used only for debugging
        MainWindow.__numInstances += 1

        self._collector = None
        self._inspector = None
        self._inspectorRegItem = None # The registered inspector item a InspectorRegItem)

        self._detailDockWidgets = []
        self._argosApplication = argosApplication
        self._configTreeModel = ConfigTreeModel()
        self._inspectorsNonDefaults = {}  # non-default values for all used plugins

        self.setCorner(Qt.TopLeftCorner, Qt.LeftDockWidgetArea)
        self.setCorner(Qt.BottomLeftCorner, Qt.LeftDockWidgetArea)
        self.setCorner(Qt.TopRightCorner, Qt.TopDockWidgetArea)
        self.setCorner(Qt.BottomRightCorner, Qt.RightDockWidgetArea)

        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.setUnifiedTitleAndToolBarOnMac(True)
        #self.setDocumentMode(True) # Look of tabs as Safari on OS-X (disabled, ugly)
        self.resize(1300, 700)  # Assumes minimal resolution of 1366 x 768

        # Move window to the center of the screen
        desktopRect = QtWidgets.QApplication.desktop().availableGeometry(self)
        center = desktopRect.center()
        self.move(center.x() - self.width () * 0.5, center.y() - self.height() * 0.5)
        
        self.__setupViews()
        self.__setupMenus()
        self.__setupDockWidgets()


    def finalize(self):
        """ Is called before destruction (when closing).
            Can be used to clean-up resources.
        """
        logger.debug("Finalizing: {}".format(self))

        # Disconnect signals
        self.collector.sigContentsChanged.disconnect(self.collectorContentsChanged)
        self._configTreeModel.sigItemChanged.disconnect(self.configContentsChanged)
        self.sigInspectorChanged.disconnect(self.inspectorSelectionPane.updateFromInspectorRegItem)
        self.customContextMenuRequested.disconnect(self.showContextMenu)


    def __setupViews(self):
        """ Creates the UI widgets.
        """
        self._collector = Collector(self.windowNumber)
        self.configWidget = ConfigWidget(self._configTreeModel)
        self.repoWidget = RepoWidget(self.argosApplication.repo, self.collector)
        # self._configTreeModel.insertItem(self.repoWidget.repoTreeView.config) # No configurable items yet

        # Define a central widget that will be the parent of the inspector widget.
        # We don't set the inspector directly as the central widget to retain the size when the
        # inspector is changed.
        widget = QtWidgets.QWidget()
        layout = QtWidgets.QVBoxLayout(widget)
        layout.setContentsMargins(CENTRAL_MARGIN, CENTRAL_MARGIN, CENTRAL_MARGIN, CENTRAL_MARGIN)
        layout.setSpacing(CENTRAL_SPACING)
        self.setCentralWidget(widget)

        # Must be after setInspector since that already draws the inspector
        self.collector.sigContentsChanged.connect(self.collectorContentsChanged)
        self._configTreeModel.sigItemChanged.connect(self.configContentsChanged)


    def __setupMenus(self):
        """ Sets up the main menu.
        """
        if True:
            # Don't use self.menuBar(), on OS-X this is not shared across windows.
            # See: http://qt-project.org/doc/qt-4.8/qmenubar.html#details
            # And:http://qt-project.org/doc/qt-4.8/qmainwindow.html#menuBar
            menuBar = QtWidgets.QMenuBar() # Make a menu without parent.
            self.setMenuBar(menuBar)
        else:
            menuBar = self.menuBar()

        ### File Menu ###

        fileMenu = menuBar.addMenu("&File")

        fileMenu.addAction("&New Window", self.cloneWindow, QtGui.QKeySequence("Ctrl+N"))
        fileMenu.addAction("Close &Window", self.close, QtGui.QKeySequence.Close)

        fileMenu.addSeparator()

        action = fileMenu.addAction("Browse Directory...",
            lambda: self.openFiles(fileMode = QtWidgets.QFileDialog.Directory))
        action.setShortcut(QtGui.QKeySequence("Ctrl+B"))

        action = fileMenu.addAction("&Open Files...",
            lambda: self.openFiles(fileMode = QtWidgets.QFileDialog.ExistingFiles))
        action.setShortcut(QtGui.QKeySequence("Ctrl+O"))

        openAsMenu = fileMenu.addMenu("Open As")
        for rtiRegItem in self.argosApplication.rtiRegistry.items:
            #rtiRegItem.tryImportClass()
            def createTrigger():
                "Function to create a closure with the regItem"
                _rtiRegItem = rtiRegItem # keep reference in closure
                return lambda: self.openFiles(rtiRegItem=_rtiRegItem,
                                              fileMode = QtWidgets.QFileDialog.ExistingFiles,
                                              caption="Open {}".format(_rtiRegItem.name))

            action = QtWidgets.QAction("{}...".format(rtiRegItem.name), self,
                enabled=True, # Since this is only executed at start-up, it must be static
                #enabled=bool(rtiRegItem.successfullyImported), # TODO: make this work?
                triggered=createTrigger())
            openAsMenu.addAction(action)

        fileMenu.addSeparator()

        # for action in self.repoWidget.repoTreeView.topLevelItemActionGroup.actions():
        #     fileMenu.addAction(action)
        #
        # for action in self.repoWidget.repoTreeView.currentItemActionGroup.actions():
        #     fileMenu.addAction(action)

        fileMenu.addSeparator()
        fileMenu.addAction("E&xit", self.argosApplication.closeAllWindows, QtGui.QKeySequence.Quit)

        ### View Menu ###
        self.viewMenu = menuBar.addMenu("&View")
        action = self.viewMenu.addAction("Installed &Plugins...", self.execPluginsDialog)
        action.setShortcut(QtGui.QKeySequence("Ctrl+P"))
        self.viewMenu.addSeparator()

        ### Inspector Menu ###
        self.execInspectorDialogAction = QtWidgets.QAction("&Browse Inspectors...", self,
                                                           triggered=self.execInspectorDialog)
        self.execInspectorDialogAction.setShortcut(QtGui.QKeySequence("Ctrl+i"))

        self.inspectorActionGroup = self.__createInspectorActionGroup(self)
        self.inspectorMenu = menuBar.addMenu("Inspector")
        addInspectorActionsToMenu(self.inspectorMenu, self.execInspectorDialogAction,
                                  self.inspectorActionGroup)

        ### Window Menu ###
        self.windowMenu = menuBar.addMenu("&Window")

        # The action added to the menu in the repopulateWindowMenu method, which is called by
        # the ArgosApplication object every time a window is added or removed.
        self.activateWindowAction = QtWidgets.QAction("Window #{}".format(self.windowNumber),
                                                      self, triggered=self.activateAndRaise,
                                                      checkable=True)
        if self.windowNumber <= 9:
            self.activateWindowAction.setShortcut(QtGui.QKeySequence("Alt+{}"
                                                                     .format(self.windowNumber)))
        ### Help Menu ###
        menuBar.addSeparator()
        helpMenu = menuBar.addMenu("&Help")
        helpMenu.addAction('&About...', self.about)

        if DEBUGGING:
            helpMenu.addSeparator()
            helpMenu.addAction("&Test", self.myTest, "Alt+T")
            helpMenu.addAction("Add Test Data", self.addTestData, "Alt+A")

        ### Context menu ###

        # Note that the dock-widgets have a Qt.PreventContextMenu context menu policy.
        # Therefor the context menu is only shown in the center widget.
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.showContextMenu)


    def __createInspectorActionGroup(self, parent):
        """ Creates an action group with 'set inspector' actions for all installed inspector.
        """
        actionGroup = QtWidgets.QActionGroup(parent)
        actionGroup.setExclusive(True)

        sortedItems = sorted(self.argosApplication.inspectorRegistry.items,
                             key=lambda item: item.identifier)
        shortCutNr = 1
        for item in sortedItems:
            logger.debug("item: {}".format(item.identifier))
            setAndDrawFn = partial(self.setAndDrawInspectorById, item.identifier)
            action = QtWidgets.QAction(item.name, self, triggered=setAndDrawFn, checkable=True)
            action.setData(item.identifier)
            if shortCutNr <= 9 and "debug" not in item.identifier: # TODO: make configurable by the user
                action.setShortcut(QtGui.QKeySequence("Ctrl+{}".format(shortCutNr)))
                shortCutNr += 1

            actionGroup.addAction(action)

        return actionGroup


    def __setupDockWidgets(self):
        """ Sets up the dock widgets. Must be called after the menu is setup.
        """
        #self.dockWidget(self.currentInspectorPane, "Current Inspector", Qt.LeftDockWidgetArea)

        self.inspectorSelectionPane = InspectorSelectionPane(self.execInspectorDialogAction,
                                                             self.inspectorActionGroup)
        self.sigInspectorChanged.connect(self.inspectorSelectionPane.updateFromInspectorRegItem)
        self.dockWidget(self.inspectorSelectionPane, "Current Inspector",
                        area=Qt.LeftDockWidgetArea)

        self.dockWidget(self.repoWidget, "Data Repository", Qt.LeftDockWidgetArea)
        self.dockWidget(self.collector, "Data Collector", Qt.TopDockWidgetArea)
        # TODO: if the title == "Settings" it won't be added to the view menu.
        self.dockWidget(self.configWidget, "Application Settings", Qt.RightDockWidgetArea)

        self.viewMenu.addSeparator()

        propertiesPane = PropertiesPane(self.repoWidget.repoTreeView)
        self.dockDetailPane(propertiesPane, area=Qt.LeftDockWidgetArea)

        attributesPane = AttributesPane(self.repoWidget.repoTreeView)
        self.dockDetailPane(attributesPane, area=Qt.LeftDockWidgetArea)

        dimensionsPane = DimensionsPane(self.repoWidget.repoTreeView)
        self.dockDetailPane(dimensionsPane, area=Qt.LeftDockWidgetArea)

        # Add am extra separator on mac because OS-X adds an 'Enter Full Screen' item
        if sys.platform.startswith('darwin'):
            self.viewMenu.addSeparator()

    ##############
    # Properties #
    ##############

    @property
    def windowNumber(self):
        """ The instance number of this window.
        """
        return self._windowNumber

    @property
    def inspectorRegItem(self):
        """ The InspectorRegItem that has been selected. Contains an InspectorRegItem.
            Can be None (e.g. at start-up).
        """
        return self._inspectorRegItem

    @property
    def inspectorId(self):
        """ The ID of the inspector registry item that has been selected.
            E.g. ''. Can be None (e.g. at start-up).
        """
        return self._inspectorRegItem.identifier if self._inspectorRegItem else None

    @property
    def inspectorName(self):
        """ The name of the inspector registry item that has been selected.
            E.g. 'Table'. Can be None (e.g. at start-up).
        """
        return self._inspectorRegItem.name if self._inspectorRegItem else None

    @property
    def inspectorFullName(self):
        """ The full name of the inspector registry item that has been selected.
            E.g. 'Qt/Table'. Can be None (e.g. at start-up).
        """
        return self._inspectorRegItem.fullName if self._inspectorRegItem else None

    @property
    def inspector(self):
        """ The inspector widget of this window. Can be None (e.g. at start-up).
        """
        return self._inspector

    @property
    def collector(self):
        """ The collector widget of this window
        """
        return self._collector

    @property
    def argosApplication(self):
        """ The ArgosApplication to which this window belongs.
        """
        return self._argosApplication

    ###########
    # Methods #
    ###########

    def repopulateWinowMenu(self, actionGroup):
        """ Clear the window menu and fills it with the actions of the actionGroup
        """
        for action in self.windowMenu.actions():
            self.windowMenu.removeAction(action)

        for action in actionGroup.actions():
            self.windowMenu.addAction(action)


    def showContextMenu(self, pos):
        """ Shows the context menu at position pos.
        """
        contextMenu = QtWidgets.QMenu()
        addInspectorActionsToMenu(contextMenu, self.execInspectorDialogAction,
                                  self.inspectorActionGroup)
        contextMenu.exec_(self.mapToGlobal(pos))


    def dockWidget(self, widget, title, area):
        """ Adds a widget as a docked widget.
            Returns the added dockWidget
        """
        assert widget.parent() is None, "Widget already has a parent"

        dockWidget = QtWidgets.QDockWidget(title, parent=self)
        dockWidget.setObjectName("dock_" + string_to_identifier(title))
        dockWidget.setWidget(widget)

        # Prevent parent context menu (with e.g. 'set inspector" option) to be displayed.
        dockWidget.setContextMenuPolicy(Qt.PreventContextMenu)

        self.addDockWidget(area, dockWidget)
        self.viewMenu.addAction(dockWidget.toggleViewAction())
        return dockWidget


    def dockDetailPane(self, detailPane, title=None, area=None):
        """ Creates a dockWidget and add the detailPane with a default title.
            By default the detail widget is added to the Qt.LeftDockWidgetArea.
        """
        title = detailPane.classLabel() if title is None else title
        area = Qt.LeftDockWidgetArea if area is None else area
        dockWidget = self.dockWidget(detailPane, title, area)
        # TODO: undockDetailPane to disconnect
        dockWidget.visibilityChanged.connect(detailPane.dockVisibilityChanged)
        if len(self._detailDockWidgets) > 0:
            self.tabifyDockWidget(self._detailDockWidgets[-1], dockWidget)
        self._detailDockWidgets.append(dockWidget)
        return dockWidget


    def updateWindowTitle(self):
        """ Updates the window title frm the window number, inspector, etc
            Also updates the Window Menu
        """
        self.setWindowTitle("{} #{} | {}-{}".format(self.inspectorName, self.windowNumber,
                                                    PROJECT_NAME, self.argosApplication.profile))
        #self.activateWindowAction.setText("{} window".format(self.inspectorName, self.windowNumber))
        self.activateWindowAction.setText("{} window".format(self.inspectorName))


    @QtSlot()
    def execInspectorDialog(self):
        """ Opens the inspector dialog box to let the user change the current inspector.
        """
        dialog = OpenInspectorDialog(self.argosApplication.inspectorRegistry, parent=self)
        dialog.setCurrentInspectorRegItem(self.inspectorRegItem)
        dialog.exec_()
        if dialog.result():
            inspectorRegItem = dialog.getCurrentInspectorRegItem()
            if inspectorRegItem is not None:
                self.getInspectorActionById(inspectorRegItem.identifier).trigger()


    def getInspectorActionById(self, identifier):
        """ Sets the inspector and draw the contents
            Triggers the corresponding action so that it is checked in the menus.
        """
        for action in self.inspectorActionGroup.actions():
            if action.data() == identifier:
                return action
        raise KeyError("No action found with ID: {!r}".format(identifier))


    def setAndDrawInspectorById(self, identifier):
        """ Sets the inspector and draw the contents.

            Does NOT trigger any actions, so the check marks in the menus are not updated. To
            achieve this, the user must update the actions by hand (or call
            getInspectorActionById(identifier).trigger() instead).
        """
        self.setInspectorById(identifier)

        # Show dialog box if import was unsuccessful.
        regItem = self.inspectorRegItem
        if regItem and not regItem.successfullyImported:
            msg = "Unable to import {} inspector.\n{}".format(regItem.identifier, regItem.exception)
            QtWidgets.QMessageBox.warning(self, "Warning", msg)
            logger.warn(msg)

        self.drawInspectorContents(reason=UpdateReason.INSPECTOR_CHANGED)


    def setInspectorById(self, identifier):
        """ Sets the central inspector widget given a inspector ID.

            If identifier is None, the inspector will be unset. Otherwise it will lookup the
            inspector class in the registry. It will raise a KeyError if the ID is not found there.

            It will do an import of the inspector code if it's loaded for the first time. If the
            the inspector class cannot be imported a warning is logged and the inspector is unset.

            NOTE: does not draw the new inspector, this is the responsibility of the caller.
            Also, the corresponding action is not triggered.

            Emits the sigInspectorChanged(self.inspectorRegItem)
        """
        logger.info("Setting inspector: {}".format(identifier))

        # Use the identifier to find a registered inspector and set self.inspectorRegItem.
        # Then create an inspector object from it.

        oldInspectorRegItem = self.inspectorRegItem
        oldInspector = self.inspector

        if not identifier:
            inspector = None
            self._inspectorRegItem = None
        else:
            inspectorRegistry = self.argosApplication.inspectorRegistry
            inspectorRegItem = inspectorRegistry.getItemById(identifier)

            self._inspectorRegItem = inspectorRegItem
            if inspectorRegItem is None:
                inspector = None
            else:
                try:
                    inspector = inspectorRegItem.create(self.collector, tryImport=True)
                except ImportError as ex:
                    # Only log the error. No dialog box or user interaction here because this
                    # function may be called at startup.
                    logger.exception("Clearing inspector. Unable to create {!r} because {}"
                                     .format(inspectorRegItem.identifier, ex))
                    inspector = None
                    self.getInspectorActionById(identifier).setEnabled(False)

                    if DEBUGGING:
                        raise

        ######################
        # Set self.inspector #
        ######################

        check_class(inspector, AbstractInspector, allow_none=True)

        logger.debug("Disabling updates.")
        self.setUpdatesEnabled(False)
        try:
            centralLayout = self.centralWidget().layout()

            # Delete old inspector
            if oldInspector is None: # can be None at start-up
                oldConfigPosition = None # Last top level element in the config tree.
            else:
                self._updateNonDefaultsForInspector(oldInspectorRegItem, oldInspector)

                # Remove old inspector configuration from tree
                oldConfigPosition = oldInspector.config.childNumber()
                configPath = oldInspector.config.nodePath
                _, oldConfigIndex = self._configTreeModel.findItemAndIndexPath(configPath)[-1]
                self._configTreeModel.deleteItemAtIndex(oldConfigIndex)

                oldInspector.finalize() # TODO: before removing config
                centralLayout.removeWidget(oldInspector)
                oldInspector.deleteLater()

            # Set new inspector
            self._inspector = inspector

            # Update collector widgets and the config tree
            oldBlockState = self.collector.blockSignals(True)
            try:
                if self.inspector is None:
                    self.collector.clearAndSetComboBoxes([])
                else:
                    # Add and apply config values to the inspector
                    key = self.inspectorRegItem.identifier
                    nonDefaults = self._inspectorsNonDefaults.get(key, {})
                    logger.debug("Setting non defaults: {}".format(nonDefaults))
                    self.inspector.config.setValuesFromDict(nonDefaults)
                    self._configTreeModel.insertItem(self.inspector.config, oldConfigPosition)
                    self.configWidget.configTreeView.expandBranch()
                    self.collector.clearAndSetComboBoxes(self.inspector.axesNames())
                    centralLayout.addWidget(self.inspector)
            finally:
                self.collector.blockSignals(oldBlockState)
        finally:
            logger.debug("Enabling updates.")
            self.setUpdatesEnabled(True)

            self.updateWindowTitle()

            logger.debug("Emitting sigInspectorChanged({})".format(self.inspectorRegItem))
            self.sigInspectorChanged.emit(self.inspectorRegItem)


    def _updateNonDefaultsForInspector(self, inspectorRegItem, inspector):
        """ Store the (non-default) config values for the current inspector in a local dictionary.
            This dictionary is later used to store value for persistence.

            This function must be called after the inspector was drawn because that may update
            some derived config values (e.g. ranges)
        """
        if inspectorRegItem and inspector:
            key = inspectorRegItem.identifier
            logger.debug("_updateNonDefaultsForInspector: {} {}"
                         .format(key, type(inspector)))
            self._inspectorsNonDefaults[key] = inspector.config.getNonDefaultsDict()
        else:
            logger.debug("_updateNonDefaultsForInspector: no inspector")


    @QtSlot()
    def execPluginsDialog(self):
        """ Shows the plugins dialog with the registered plugins
        """
        pluginsDialog = PluginsDialog(parent=self,
                                inspectorRegistry=self.argosApplication.inspectorRegistry,
                                rtiRegistry=self.argosApplication.rtiRegistry)
        pluginsDialog.exec_()


    @QtSlot(str)
    def collectorContentsChanged(self, reason):
        """ Slot that updates the UI whenever the contents of the collector has changed.
        """
        logger.debug("collectorContentsChanged()")
        self.drawInspectorContents(reason=reason)


    @QtSlot(AbstractCti)
    def configContentsChanged(self, configTreeItem):
        """ Slot is called when an item has been changed by setData of the ConfigTreeModel.
            Will draw the window contents.
        """
        logger.debug("configContentsChanged: {}".format(configTreeItem))
        self.drawInspectorContents(reason=UpdateReason.CONFIG_CHANGED,
                                   origin=configTreeItem)


    def drawInspectorContents(self, reason, origin=None):
        """ Draws all contents of this window's inspector.
            The reason and origin parameters are passed on to the inspector's updateContents method.

            :param reason: string describing the reason for the redraw.
                Should preferably be one of the UpdateReason enumeration class, but new values may
                be used (which are then ignored by existing inspectors).
            :param origin: object with extra infor on the reason
        """
        logger.debug("")
        logger.debug("-------- Drawing inspector of window: {} --------".format(self.windowTitle()))
        if self.inspector:
            self.inspector.updateContents(reason=reason, initiator=origin)
        else:
            logger.debug("No inspector selected")
        logger.debug("Finished draw inspector.\n")


    # TODO: to repotreemodel? Note that the functionality will be common to selectors.
    @QtSlot()
    def openFiles(self, fileNames=None, rtiRegItem=None, caption=None, fileMode=None):
        """ Lets the user select on or more files and opens it.

            :param fileNames: If None an open-file dialog allows the user to select files,
                otherwise the files are opened directly.
            :param rtiRegItem: Open the files as this type of registered RTI. None=autodetect.
            :param caption: Optional caption for the file dialog.
            :param fileMode: is passed to the file dialog.
            :rtype fileMode: QtWidgets.QFileDialog.FileMode constant
        """
        if fileNames is None:
            dialog = QtWidgets.QFileDialog(self, caption=caption)

            if rtiRegItem is None:
                nameFilter = 'All files (*);;' # Default show all files.
                nameFilter += self.argosApplication.rtiRegistry.getFileDialogFilter()
            else:
                nameFilter = rtiRegItem.getFileDialogFilter()
                nameFilter += ';;All files (*)'
            dialog.setNameFilter(nameFilter)

            if fileMode:
                dialog.setFileMode(fileMode)

            if dialog.exec_() == QtWidgets.QFileDialog.Accepted:
                fileNames = dialog.selectedFiles()
            else:
                fileNames = []

        fileRootIndex = None
        for fileName in fileNames:
            rtiClass = rtiRegItem.getClass(tryImport=True) if rtiRegItem else None
            fileRootIndex = self.argosApplication.repo.loadFile(fileName, rtiClass=rtiClass)
            self.repoWidget.repoTreeView.setExpanded(fileRootIndex, True)

        # Select last opened file
        if fileRootIndex is not None:
            self.repoWidget.repoTreeView.setCurrentIndex(fileRootIndex)


    def trySelectRtiByPath(self, path):
        """ Selects a repository tree item given a path, expanding nodes if along the way if needed.

            Returns (item, index) if the path was selected successfully, else a warning is logged
            and (None, None) is returned.
        """
        try:
            lastItem, lastIndex = self.repoWidget.repoTreeView.expandPath(path)
            self.repoWidget.repoTreeView.setCurrentIndex(lastIndex)
            return lastItem, lastIndex
        except Exception as ex:
            logger.warn("Unable to select {!r} because: {}".format(path, ex))
            if DEBUGGING:
                raise
            return None, None


    def readViewSettings(self, settings=None): # TODO: rename to readProfile?
        """ Reads the persistent program settings

            :param settings: optional QSettings object which can have a group already opened.
            :returns: True if the header state was restored, otherwise returns False
        """
        if settings is None:
            settings = QtCore.QSettings()
        logger.debug("Reading settings from: {}".format(settings.group()))

        self.restoreGeometry(settings.value("geometry"))
        self.restoreState(settings.value("state"))

        self.repoWidget.repoTreeView.readViewSettings('repo_tree/header_state', settings)
        self.configWidget.configTreeView.readViewSettings('config_tree/header_state', settings)

        #self._configTreeModel.readModelSettings('config_model', settings)
        settings.beginGroup('cfg_inspectors')
        try:
            for key in settings.childKeys():
                json = settings.value(key)
                self._inspectorsNonDefaults[key] = ctiLoads(json)
        finally:
            settings.endGroup()

        identifier = settings.value("inspector", None)
        try:
            if identifier:
                self.setInspectorById(identifier)
        except KeyError as ex:
            logger.warn("No inspector with ID {!r}.: {}".format(identifier, ex))


    def saveProfile(self, settings=None):
        """ Writes the view settings to the persistent store
        """
        self._updateNonDefaultsForInspector(self.inspectorRegItem, self.inspector)

        if settings is None:
            settings = QtCore.QSettings()
        logger.debug("Writing settings to: {}".format(settings.group()))

        settings.beginGroup('cfg_inspectors')
        try:
            for key, nonDefaults in self._inspectorsNonDefaults.items():
                if nonDefaults:
                    settings.setValue(key, ctiDumps(nonDefaults))
                    logger.debug("Writing non defaults for {}: {}".format(key, nonDefaults))
        finally:
            settings.endGroup()

        self.configWidget.configTreeView.saveProfile("config_tree/header_state", settings)
        self.repoWidget.repoTreeView.saveProfile("repo_tree/header_state", settings)

        settings.setValue("geometry", self.saveGeometry())
        settings.setValue("state", self.saveState())

        identifier = self.inspectorRegItem.identifier if self.inspectorRegItem else ''
        settings.setValue("inspector", identifier)


    @QtSlot()
    def cloneWindow(self):
        """ Opens a new window with the same inspector as the current window.
        """
        # Save current window settings.
        settings = QtCore.QSettings()
        settings.beginGroup(self.argosApplication.windowGroupName(self.windowNumber))
        try:
            self.saveProfile(settings)

            # Create new window with the freshly baked settings of the current window.
            name = self.inspectorRegItem.fullName
            newWindow = self.argosApplication.addNewMainWindow(settings=settings,
                                                               inspectorFullName=name)
        finally:
            settings.endGroup()

        # Select the current item in the new window.
        currentItem, _currentIndex = self.repoWidget.repoTreeView.getCurrentItem()
        if currentItem:
            newWindow.trySelectRtiByPath(currentItem.nodePath)

        # Move the new window 24 pixels to the bottom right and raise it to the front.
        newGeomRect = newWindow.geometry()
        logger.debug("newGeomRect: x={}".format(newGeomRect.x()))
        newGeomRect.moveTo(newGeomRect.x() + 24, newGeomRect.y() + 24)

        newWindow.setGeometry(newGeomRect)
        logger.debug("newGeomRect: x={}".format(newGeomRect.x()))

        newWindow.raise_()


    @QtSlot()
    def activateAndRaise(self):
        """ Activates and raises the window.
        """
        logger.debug("Activate and raising window: {}".format(self.windowNumber))
        self.activateWindow()
        self.raise_()


    def event(self, ev):
        """ Detects the WindowActivate event. Pass all event through to the super class.
        """
        if ev.type() == QtCore.QEvent.WindowActivate:
            logger.debug("Window activated: {}".format(self.windowNumber))
            self.activateWindowAction.setChecked(True)

        return super(MainWindow, self).event(ev);


    def closeEvent(self, event):
        """ Called when closing this window.
        """
        logger.debug("closeEvent")
        self.argosApplication.saveSettingsIfNeeded()
        self.finalize()
        self.argosApplication.removeMainWindow(self)
        event.accept()
        logger.debug("closeEvent accepted")


    @QtSlot()
    def about(self):
        """ Shows the about message window.
        """
        aboutDialog = AboutDialog(parent=self)
        aboutDialog.show()
        aboutDialog.addDependencyInfo()


    @QtSlot()
    def addTestData(self):
        """ Adds test data to the repository
        """
        logger.info("Adding test data to the repository.")
        self.argosApplication.repo.insertItem(createArgosTestData())


    @QtSlot()
    def myTest(self):
        """ Function for small ad-hoc tests
        """
        logger.info("myTest function called")
        self.testSelectAllData()

        # self.inspector.config.setAutoRangeOn(2)
        # self.inspector.config.setColorAutoRangeOn()
        # self.inspector.config.histRangeCti.setAutoRangeOn()

        # logger.debug("Repo icon size: {}".format(self.repoWidget.repoTreeView.iconSize()))
        # #self.repoWidget.repoTreeView.setIconSize(QtCore.QSize(32, 32))
        # self.repoWidget.repoTreeView.setIconSize(QtCore.QSize(24, 24))
        # logger.debug("Repo icon size: {}".format(self.repoWidget.repoTreeView.iconSize()))
        # #self.collector.tree.resizeColumnsToContents(startCol=1)

        # from argos.qt.misc import printChildren
        # printChildren(self.centralWidget())
        # print()
        # print()
        #
        #
        # self.argosApplication.raiseAllWindows()
        # import gc
        # from argos.qt import printAllWidgets
        # printAllWidgets(self._argosApplication._qApplication, ofType=MainWindow)
        # print("forcing garbage collection")
        # gc.collect()
        # printAllWidgets(self._argosApplication._qApplication, ofType=MainWindow)

        #for item in self.argosApplication.inspectorRegistry.items:
        #    logger.debug("item: {}".format(item))


    def testSelectAllData(self):
        """ Selects all nodes in a subtree for all inspectors
        """
        # Skip nodes that give known, unfixable errors.
        skipPaths = ['/myDict/numbers/-inf', '/myDict/numbers/nan',
                     '/myDict/numbers/large float'] # in 2D image plot

        def visitNodes(index):
            """ Visits all the nodes recursively.
            """
            assert index.isValid(), "sanity check"

            repoModel = self.repoWidget.repoTreeView.model()
            item = repoModel.getItem(index)
            logger.info("Visiting: {!r} ({} children)".
                        format(item.nodePath, repoModel.rowCount(index)))

            # Select index
            if item.nodePath in skipPaths:
                logger.warn("Skipping node during testing: {}".format(item.nodePath))
                return

            self.repoWidget.repoTreeView.setCurrentIndex(index)
            QtWidgets.qApp.processEvents() # Cause Qt to update UI

            # Expand node to load children.
            #self.repoWidget.repoTreeView.setExpanded(index, True)
            #QtWidgets.qApp.processEvents() # Cause Qt to load children.

            for rowNr in range(repoModel.rowCount(index)):
                childIndex = repoModel.index(rowNr, 0, parentIndex=index)
                visitNodes(childIndex)

        # Actual boddy
        rootNodes = ['/myDict']
        #rootNodes = ['/argos/icm/S5P_ICM_CA_UVN_20120919T051721_20120919T065655_01890_01_001000_20151002T140000.h5']

        for rootNode in rootNodes:
            logger.info("Selecting all nodes in: {}".format(rootNode))

            nodeItem, nodeIndex = self.trySelectRtiByPath(rootNode)
            self.repoWidget.repoTreeView.expandBranch(index = nodeIndex, expanded=True) # TODO: why necessary?
            #QtWidgets.qApp.processEvents()
            visitNodes(nodeIndex)

            #
            # try:
            #
            #     self.repoWidget.repoTreeView.setCurrentIndex(lastIndex)
            # except Exception as ex:
            #     logger.error(ex)
            #     if DEBUGGING:
            #         raise




# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Defines a global RTI registry to register repository tree item plugins.
"""

import logging
from argos.utils.cls import check_is_a_string, check_class, check_is_a_sequence
from argos.utils.misc import prepend_point_to_extension
from argos.qt.registry import ClassRegItem, ClassRegistry, GRP_REGISTRY

logger = logging.getLogger(__name__)

GRP_REGISTRY_RTI = GRP_REGISTRY + '/rti'

class RtiRegItem(ClassRegItem):
    """ Class to keep track of a registered Repo Tree Item.
    """
    def __init__(self, fullName, fullClassName, extensions, pythonPath=''):
        """ Constructor. See the ClassRegItem class doc string for the parameter help.
        """
        super(RtiRegItem, self).__init__(fullName, fullClassName, pythonPath=pythonPath)
        self._extensions = [prepend_point_to_extension(ext) for ext in extensions]


    @property
    def extensions(self):
        """ Filename extensions that will automatically open as this RTI.
        """
        return self._extensions


    def getFileDialogFilter(self):
        """ Returns a filters that can be used to construct file dialogs filters,
            for example: 'Text File (*.txt;*.text)'
        """
        extStr = ';'.join(['*' + ext for ext in self.extensions])
        return '{} ({})'.format(self.name, extStr)


    def asDict(self):
        """ Returns a dictionary for serialization.
        """
        dct = super(RtiRegItem, self).asDict()
        dct['extensions'] = self.extensions
        return dct


class RtiRegistry(ClassRegistry):
    """ Class that can be used to register repository tree items (RTIs).

        Maintains a name to RtiClass mapping and an extension to RtiClass mapping.
        The extension in the extensionToRti assure that a unique RTI is used as default mapping,
        the extensions in the RtiRegItem class do not have to be unique and are used in the
        filter in the getFileDialogFilter function.
    """
    def __init__(self, settingsGroupName=GRP_REGISTRY_RTI):
        """ Constructor
        """
        super(RtiRegistry, self).__init__(settingsGroupName=settingsGroupName)
        self._itemClass = RtiRegItem
        self._extensionMap = {}


    def clear(self):
        """ Empties the registry
        """
        super(RtiRegistry, self).clear()
        self._extensionMap = {}


    def _registerExtension(self, extension, rtiRegItem):
        """ Links an file name extension to a repository tree item.
        """
        check_is_a_string(extension)
        check_class(rtiRegItem, RtiRegItem)

        logger.debug("  Registering extension {!r} for {}".format(extension, rtiRegItem))

        # TODO: type checking
        if extension in self._extensionMap:
            logger.warn("Overriding extension {!r}: old={}, new={}"
                        .format(extension, self._extensionMap[extension], rtiRegItem))
        self._extensionMap[extension] = rtiRegItem


    def registerItem(self, regItem):
        """ Adds a ClassRegItem object to the registry.
        """
        super(RtiRegistry, self).registerItem(regItem)

        for ext in regItem.extensions:
            self._registerExtension(ext, regItem)


    def registerRti(self, fullName, fullClassName, extensions=None, pythonPath=''):
        """ Class that maintains the collection of registered inspector classes.
            Maintains a lit of file extensions that open the RTI by default.
        """
        check_is_a_sequence(extensions)
        extensions = extensions if extensions is not None else []
        extensions = [prepend_point_to_extension(ext) for ext in extensions]

        regRti = RtiRegItem(fullName, fullClassName, extensions, pythonPath=pythonPath)
        self.registerItem(regRti)


    def getRtiRegItemByExtension(self, extension):
        """ Returns the RtiRegItem class registered for the extension.
            Raise KeyError if no class registered for the extension.
        """
        rtiRegItem = self._extensionMap[extension]
        return rtiRegItem


    def getFileDialogFilter(self):
        """ Returns a filter that can be used in open file dialogs,
            for example: 'All files (*);;Txt (*.txt;*.text);;netCDF(*.nc;*.nc4)'
        """
        filters = []
        for regRti in self.items:
            filters.append(regRti.getFileDialogFilter())
        return ';;'.join(filters)


    def getDefaultItems(self):
        """ Returns a list with the default plugins in the repo tree item registry.
        """
        return [
            RtiRegItem('HDF-5 file',
                       'argos.repo.rtiplugins.hdf5.H5pyFileRti',
                       extensions=['hdf5', 'h5', 'h5e', 'he5', 'nc']), # hdf extension is for HDF-4

            RtiRegItem('MATLAB file',
                       'argos.repo.rtiplugins.scipyio.MatlabFileRti',
                       extensions=['mat']),

            RtiRegItem('NetCDF file',
                       'argos.repo.rtiplugins.ncdf.NcdfFileRti',
                       #extensions=['nc', 'nc3', 'nc4']),
                       extensions=['nc', 'nc4']),
                       #extensions=[]),

            RtiRegItem('NumPy binary file',
                       'argos.repo.rtiplugins.numpyio.NumpyBinaryFileRti',
                       extensions=['npy']),

            RtiRegItem('NumPy compressed file',
                       'argos.repo.rtiplugins.numpyio.NumpyCompressedFileRti',
                       extensions=['npz']),

            RtiRegItem('NumPy text file',
                       'argos.repo.rtiplugins.numpyio.NumpyTextFileRti',
                       #extensions=['txt', 'text']),
                       extensions=['dat']),

            RtiRegItem('IDL save file',
                       'argos.repo.rtiplugins.scipyio.IdlSaveFileRti',
                       extensions=['sav']),

            RtiRegItem('Pandas CSV file',
                       'argos.repo.rtiplugins.pandasio.PandasCsvFileRti',
                        extensions=['csv']),

            RtiRegItem('Pillow image',
                       'argos.repo.rtiplugins.pillowio.PillowFileRti',
                        extensions=['bmp', 'eps', 'im', 'gif', 'jpg', 'jpeg', 'msp', 'pcx',
                                    'png', 'ppm', 'spi', 'tif', 'tiff', 'xbm', 'xv']),

            RtiRegItem('Wav file',
                       'argos.repo.rtiplugins.scipyio.WavFileRti',
                       extensions=['wav'])]


# The RTI registry is implemented as a singleton. This is necessary because
# in DirectoryRti._fetchAllChildren we need access to the registry.
# TODO: think of an elegant way to access the ArgosApplication.registry from there.
def createGlobalRegistryFunction():
    """ Closure to create the RtiRegistry singleton
    """
    globReg = RtiRegistry()

    def accessGlobalRegistry():
        return globReg

    return accessGlobalRegistry

# This is actually a function definition, not a constant
#pylint: disable=C0103

globalRtiRegistry = createGlobalRegistryFunction()
globalRtiRegistry.__doc__ = "Function that returns the RtiRegistry singleton common to all windows"






# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Repository items (RTIs) for browsing the file system
"""

import logging, os
from argos.repo.baserti import BaseRti
from argos.qt import QtWidgets
from argos.repo.iconfactory import RtiIconFactory
from argos.repo.registry import globalRtiRegistry

logger = logging.getLogger(__name__)


class UnknownFileRti(BaseRti):
    """ A repository tree item that represents a file of unknown type.
        The file is not opened.
    """
    _defaultIconGlyph = RtiIconFactory.FILE
    _defaultIconColor = RtiIconFactory.COLOR_UNKNOWN

    def __init__(self, nodeName='', fileName=''):
        """ Constructor
        """
        super(UnknownFileRti, self).__init__(nodeName=nodeName, fileName=fileName)
        self._checkFileExists()


    def hasChildren(self):
        """ Returns False. Leaf nodes never have children. """
        return False



class DirectoryRti(BaseRti):
    """ A repository tree item that has a reference to a file.
    """
    _defaultIconGlyph = RtiIconFactory.FOLDER
    _defaultIconColor = RtiIconFactory.COLOR_UNKNOWN

    def __init__(self, nodeName='', fileName=''):
        """ Constructor
        """
        super(DirectoryRti, self).__init__(nodeName=nodeName, fileName=fileName)
        self._checkFileExists() # TODO: check for directory?


    def _fetchAllChildren(self):
        """ Gets all sub directories and files within the current directory.
            Does not fetch hidden files.
        """
        childItems = []
        fileNames = os.listdir(self._fileName)
        absFileNames = [os.path.join(self._fileName, fn) for fn in fileNames]

        # Add subdirectories
        for fileName, absFileName in zip(fileNames, absFileNames):
            if os.path.isdir(absFileName) and not fileName.startswith('.'):
                childItems.append(DirectoryRti(fileName=absFileName, nodeName=fileName))

        # Add regular files
        for fileName, absFileName in zip(fileNames, absFileNames):
            if os.path.isfile(absFileName) and not fileName.startswith('.'):
                childItem = createRtiFromFileName(absFileName)
                childItems.append(childItem)

        return childItems


def detectRtiFromFileName(fileName):
    """ Determines the type of RepoTreeItem to use given a file name.
        Uses a DirectoryRti for directories and an UnknownFileRti if the file
        extension doesn't match one of the registered RTI extensions.

        Returns (cls, regItem) tuple. Both the cls ond the regItem can be None.
        If the file is a directory, (DirectoryRti, None) is returned.
        If the file extension is not in the registry, (UnknownFileRti, None) is returned.
        If the cls cannot be imported (None, regItem) returned. regItem.exception will be set.
        Otherwise (cls, regItem) will be returned.
    """
    _, extension = os.path.splitext(fileName)
    if os.path.isdir(fileName):
        rtiRegItem = None
        cls = DirectoryRti
    else:
        try:
            rtiRegItem = globalRtiRegistry().getRtiRegItemByExtension(extension)
        except (KeyError):
            logger.debug("No file RTI registered for extension: {}".format(extension))
            rtiRegItem = None
            cls = UnknownFileRti
        else:
            cls = rtiRegItem.getClass(tryImport=True) # cls can be None

    return cls, rtiRegItem


def createRtiFromFileName(fileName):
    """ Determines the type of RepoTreeItem to use given a file name and creates it.
        Uses a DirectoryRti for directories and an UnknownFileRti if the file
        extension doesn't match one of the registered RTI extensions.
    """
    cls, rtiRegItem = detectRtiFromFileName(fileName)
    if cls is None:
        logger.warn("Unable to import plugin {}: {}"
                    .format(rtiRegItem.fullName, rtiRegItem.exception))
        rti = UnknownFileRti.createFromFileName(fileName)
        rti.setException(rtiRegItem.exception)
    else:
        rti = cls.createFromFileName(fileName)

    assert rti, "Sanity check failed (createRtiFromFileName). Please report this bug."
    return rti




"""
    'About' dialog window that shows version information.

"""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import logging

from argos.info import PROJECT_NAME, VERSION
from argos.qt import QtCore, QtGui, QtWidgets
from argos.utils.cls import is_a_string
from argos.utils import moduleinfo as mi
from argos.widgets.constants import MONO_FONT, FONT_SIZE

logger = logging.getLogger(__name__)


# The main window inherits from a Qt class, therefore it has many
# ancestors public methods and attributes.
# pylint: disable=R0901, R0902, R0904, W0201


class AboutDialog(QtWidgets.QDialog):
    """ Dialog window that shows dependency information.
    """
    def __init__(self, parent=None):
        """ Constructor
        """
        super(AboutDialog, self).__init__(parent=parent)
        self.setModal(True)

        mainLayout = QtWidgets.QVBoxLayout()
        self.setLayout(mainLayout)

        progVersionLabel = QtWidgets.QLabel()
        progVersionLabel.setText("{} {}".format(PROJECT_NAME, VERSION))
        progVersionLabel.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        mainLayout.addWidget(progVersionLabel)

        font = QtGui.QFont()
        font.setFamily(MONO_FONT)
        font.setFixedPitch(True)
        font.setPointSize(FONT_SIZE)

        self.editor = QtWidgets.QPlainTextEdit()
        self.editor.setReadOnly(True)
        self.editor.setFont(font)
        self.editor.setWordWrapMode(QtGui.QTextOption.NoWrap)
        self.editor.clear()
        self.editor.setPlainText("Retrieving package info...")
        mainLayout.addWidget(self.editor)

        self.progressLabel = QtWidgets.QLabel()
        mainLayout.addWidget(self.progressLabel)

        buttonBox = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok)
        buttonBox.accepted.connect(self.accept)
        mainLayout.addWidget(buttonBox)

        self.resize(QtCore.QSize(800, 400))


    def _addModuleInfo(self, moduleInfo):
        """ Adds a line with module info to the editor
            :param moduleInfo: can either be a string or a module info class.
                In the first case, an object is instantiated as ImportedModuleInfo(moduleInfo).
        """
        if is_a_string(moduleInfo):
            moduleInfo = mi.ImportedModuleInfo(moduleInfo)

        line = "{:15s}: {}".format(moduleInfo.name, moduleInfo.verboseVersion)
        self.editor.appendPlainText(line)
        QtWidgets.QApplication.instance().processEvents()


    def addDependencyInfo(self):
        """ Adds version info about the installed dependencies
        """
        logger.debug("Adding dependency info to the AboutDialog")
        self.progressLabel.setText("Retrieving package info...")
        self.editor.clear()

        self._addModuleInfo(mi.PythonModuleInfo())
        self._addModuleInfo(mi.QtModuleInfo())

        modules = ['numpy', 'scipy', 'pandas', 'pyqtgraph']
        for module in modules:
            self._addModuleInfo(module)

        self._addModuleInfo(mi.PillowInfo())
        self._addModuleInfo(mi.H5pyModuleInfo())
        self._addModuleInfo(mi.NetCDF4ModuleInfo())

        self.progressLabel.setText("")
        logger.debug("Finished adding dependency info to the AboutDialog")




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos.  If not, see <http://www.gnu.org/licenses/>.

""" Argos numerical data inspector.
"""
from __future__ import print_function

# ----IMPORTANT ----
# Do not do any imports here that (indirectly) import any dependencies (PyQt, numpy, etc)
# The browse function is imported by the argos package, which in turn is imported by setup.py.
# If you import (for instance) numpy here, the setup.py will fail if numpy is not installed.

import logging, sys, argparse
from argos.info import DEBUGGING, PROJECT_NAME, VERSION, DEFAULT_PROFILE

logger = logging.getLogger('argos')

logging.basicConfig(level='DEBUG', stream=sys.stderr,
                    #format='%(name)35s %(asctime)s %(filename)25s:%(lineno)-4d : %(levelname)-7s: %(message)s')
                    format='%(asctime)s %(filename)25s:%(lineno)-4d : %(levelname)-7s: %(message)s')


def browse(fileNames=None,
           inspectorFullName=None,
           select=None,
           profile=DEFAULT_PROFILE,
           resetProfile=False,      # TODO: should probably be moved to the main program
           resetAllProfiles=False,  # TODO: should probably be moved to the main program
           resetRegistry=False):    # TODO: should probably be moved to the main program
    """ Opens the main window(s) for the persistent settings of the given profile,
        and executes the application.

        :param fileNames: List of file names that will be added to the repository
        :param inspectorFullName: The full path name of the inspector that will be loaded
        :param select: a path of the repository item that will selected at start up.
        :param profile: the name of the profile that will be loaded
        :param resetProfile: if True, the profile will be reset to it standard settings.
        :param resetAllProfiles: if True, all profiles will be reset to it standard settings.
        :param resetRegistry: if True, the registry will be reset to it standard settings.
        :return:
    """
    # Imported here so this module can be imported without Qt being installed.
    from argos.qt import QtWidgets, QtCore
    from argos.application import ArgosApplication
    from argos.repo.testdata import createArgosTestData

    try:
        QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps)
    except Exception as ex:
        logger.debug("AA_UseHighDpiPixmaps not available in PyQt4: {}".format(ex))

    # Create
    argosApp = ArgosApplication()

    if resetProfile:
        argosApp.deleteProfile(profile)
    if resetAllProfiles:
        argosApp.deleteAllProfiles()
    if resetRegistry:
        argosApp.deleteRegistries()

    # Must be called before opening the files so that file formats are auto-detected.
    argosApp.loadOrInitRegistries()

    # Load data in common repository before windows are created.
    argosApp.loadFiles(fileNames)
    if DEBUGGING:
        argosApp.repo.insertItem(createArgosTestData())

    # Create windows for this profile.
    argosApp.loadProfile(profile=profile, inspectorFullName=inspectorFullName)

    if select:
        for mainWindow in argosApp.mainWindows:
            mainWindow.trySelectRtiByPath(select)

    return argosApp.execute()


def printInspectors():
    """ Prints a list of inspectors
    """
    # Imported here so this module can be imported without Qt being installed.
    from argos.application import ArgosApplication

    argosApp = ArgosApplication()
    argosApp.loadOrInitRegistries()
    for regItem in argosApp.inspectorRegistry.items:
        print(regItem.fullName)

# TODO: better logging
# def configBasicLogging(level = 'DEBUG'):
#     """ Setup basic config logging.
#     """
#     fmt = '%(name)40s %(asctime)s %(filename)25s:%(lineno)-4d : %(levelname)-7s: %(message)s'
#     logging.basicConfig(level=level, format=fmt)


def remove_process_serial_number(arg_list):
    """ Creates a copy of a list (typically sys.argv) where the strings that
        start with '-psn_0_' are removed.

        These are the process serial number used by the OS-X open command
        to bring applications to the front. They clash with argparse.
        See: http://hintsforums.macworld.com/showthread.php?t=11978
    """
    return [arg for arg in arg_list if not arg.startswith("-psn_0_")]


def main():
    """ Starts Argos main window
    """
    about_str = "{} version: {}".format(PROJECT_NAME, VERSION)
    parser = argparse.ArgumentParser(description = about_str)

    parser.add_argument('fileNames', metavar='FILE', nargs='*', help='Input files')

    parser.add_argument('-i', '--inspector', dest='inspector',
        help="""The identifier or fullName of the inspector that will be opened at start up.
                E.g. 'Qt/Table'""")

    parser.add_argument('--list-inspectors', dest='list_inspectors', action = 'store_true',
        help="""Prints a list of available inspectors for the -i option""")


    parser.add_argument('-s', '--select', dest='selectPath',
        help="""Full path name of a repository tree item that will be selected at start-up.
                E.g. 'file/var/fieldname'""")

    parser.add_argument('-p', '--profile', dest='profile', default=DEFAULT_PROFILE,
        help="Can be used to have different persistent settings for different use cases.")

    parser.add_argument('--reset', '--reset-profile', dest='reset_profile', action = 'store_true',
        help="If set, persistent settings will be reset for the current profile.")

    parser.add_argument('--reset-all-profiles', dest='reset_all_profiles', action = 'store_true',
        help="If set, persistent settings will be reset for the all profiles.")

    parser.add_argument('--reset-registry', dest='reset_registry', action = 'store_true',
        help="If set, the registry will be reset to contain only the default plugins.")

    parser.add_argument('--version', action = 'store_true',
        help="Prints the program version.")

    parser.add_argument('-l', '--log-level', dest='log_level', default='warning',
        help="Log level. Only log messages with a level higher or equal than this will be printed. "
             "Default: 'warning'",
        choices=('debug', 'info', 'warning', 'error', 'critical'))

    args = parser.parse_args(remove_process_serial_number(sys.argv[1:]))

    if DEBUGGING:
        logger.info("Setting log level to: {}".format(args.log_level.upper()))
    logger.setLevel(args.log_level.upper())

    if DEBUGGING:
        logger.warn("Debugging flag is on!")

    if args.version:
        print(about_str)
        sys.exit(0)

    if args.list_inspectors:
        printInspectors()
        sys.exit(0)

    logger.info('Started {} {}'.format(PROJECT_NAME, VERSION))
    logger.info("Python version: {}".format(sys.version).replace('\n', ''))
    #logger.info('Using: {}'.format('PyQt' if USE_PYQT else 'PySide'))

    # Imported here so this module can be imported without Qt being installed.
    from argos.qt.misc import ABOUT_QT_BINDINGS
    logger.info("Using {}".format(ABOUT_QT_BINDINGS))

    # Browse will create an ArgosApplication with one MainWindow
    browse(fileNames = args.fileNames,
           inspectorFullName=args.inspector,
           select=args.selectPath,
           profile=args.profile,
           resetProfile=args.reset_profile,
           resetAllProfiles=args.reset_all_profiles,
           resetRegistry=args.reset_registry)
    logger.info('Done {}'.format(PROJECT_NAME))

if __name__ == "__main__":
    main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-


import unittest, logging, sys
import numpy as np

from numpy.testing import assert_array_equal
from argos import configBasicLogging
from argos.repo.memoryrtis import ArrayRti



class TestUntypedCtis(unittest.TestCase):

    def setUp(self):
        self.arr = np.arange(24).reshape(6, 4)
        self.rti = ArrayRti(self.arr, 'my array')


    def test_indexing(self):
        """ Test indexing and slicing
        """

        self.assertEqual(self.rti[2, 1], self.arr[2, 1])
        assert_array_equal(self.rti[:], self.arr[:])
        assert_array_equal(self.rti[2:5, :], self.arr[2:5, :])
        assert_array_equal(self.rti[-1], self.arr[-1])

        # Multidimensional indexing
        idx0 = [1, 1, 3]
        idx1 = [0, -1, 2]
        assert_array_equal(self.rti[idx0, idx1], self.arr[idx0, idx1])

        slices = [slice(2, None), 1]
        assert_array_equal(self.rti[slices], self.arr[slices])

        slices = tuple([slice(2), slice(1)])
        assert_array_equal(self.rti[slices], self.arr[slices])



if __name__ == '__main__':
    configBasicLogging(level='DEBUG')
    unittest.main()





# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Repository Tree Items (RTIs) for HDF-5 data.

    It uses the h5py package to open HDF-5 files.
    See http://www.h5py.org/
"""

import logging, os
import h5py
import numpy as np


from argos.repo.iconfactory import RtiIconFactory
from argos.repo.baserti import BaseRti
from argos.utils.cls import to_string, check_class, is_an_array
from argos.utils.masks import maskedEqual

logger = logging.getLogger(__name__)

ICON_COLOR_H5PY = '#00EE88'


def dimNamesFromDataset(h5Dataset):
    """ Constructs the dimension names given a h5py dataset.

        First looks in the dataset's dimension scales to see if it refers to another
        dataset. In that case the referred dataset's name is used. If not, the label of the
        dimension scale is used. Finally, if this is empty, the dimension is numbered.
    """
    dimNames = [] # TODO: cache?
    for dimNr, dimScales in enumerate(h5Dataset.dims):
        if len(dimScales) == 0:
            dimNames.append('Dim{}'.format(dimNr))
        elif len(dimScales) == 1:
            dimScaleLabel, dimScaleDataset = dimScales.items()[0]
            path = dimScaleDataset.name
            if path:
                dimNames.append(os.path.basename(path))
            elif dimScaleLabel: # This could potentially be long so it's our second choice
                dimNames.append(dimScaleLabel)
            else:
                dimNames.append('Dim{}'.format(dimNr))
        else:
            # TODO: multiple scales for this dimension. What to do?
            logger.warn("More than one dimension scale found: {!r}".format(dimScales))
            dimNames.append('Dim{}'.format(dimNr)) # For now, just number them

    return dimNames


def dataSetElementType(h5Dataset):
    """ Returns a string describing the element type of the dataset
    """
    dtype =  h5Dataset.dtype

    if dtype.names:
        return '<structured>'
    else:
        if dtype.metadata and 'vlen' in dtype.metadata:
            vlen_type = dtype.metadata['vlen']
            try:
                return "<vlen {}>".format(vlen_type.__name__)  # when vlen_type is a type
            except AttributeError: #
                return "<vlen {}>".format(vlen_type.name)      # when vlen_type is a dtype

    return str(dtype)


def dataSetUnit(h5Dataset):
    """ Returns the unit of the h5Dataset by looking in the attributes.

        It searches in the attributes for one of the following keys:
        'unit', 'units', 'Unit', 'Units', 'UNIT', 'UNITS'. If these are not found, the empty
        string is returned.

        Always returns a string
    """
    attributes = h5Dataset.attrs
    if not attributes:
        return '' # a premature optimization :-)

    for key in ('unit', 'units', 'Unit', 'Units', 'UNIT', 'UNITS'):
        if key in attributes:
            # In Python3 the attributes are byte strings so we must decode them
            # This a bug in h5py, see https://github.com/h5py/h5py/issues/379
            return to_string(attributes[key])
    # Not found
    return ''


def dataSetMissingValue(h5Dataset):
    """ Returns the missingData given a HDF-5 dataset

        Looks for one of the following attributes: _FillValue, missing_value, MissingValue,
        missingValue. Returns None if these attributes are not found.

        HDF-EOS and NetCDF files seem to put the attributes in 1-element arrays. So if the
        attribute contains an array of one element, that first element is returned here.
    """
    attributes = h5Dataset.attrs
    if not attributes:
        return None # a premature optimization :-)

    for key in ('missing_value', 'MissingValue', 'missingValue', 'FillValue', '_FillValue'):
        if key in attributes:
            missingDataValue = attributes[key]
            if is_an_array(missingDataValue) and len(missingDataValue) == 1:
                return missingDataValue[0] # In case of HDF-EOS and NetCDF files
            else:
                return missingDataValue
    return None



class H5pyScalarRti(BaseRti):
    """ Repository Tree Item (RTI) that contains a scalar HDF-5 variable.

    """
    _defaultIconGlyph = RtiIconFactory.SCALAR
    _defaultIconColor = ICON_COLOR_H5PY

    def __init__(self, h5Dataset, nodeName='', fileName=''):
        """ Constructor
        """
        super(H5pyScalarRti, self).__init__(nodeName = nodeName, fileName=fileName)
        check_class(h5Dataset, h5py.Dataset)
        self._h5Dataset = h5Dataset


    def hasChildren(self):
        """ Returns False. Leaf nodes never have children. """
        return False


    @property
    def isSliceable(self):
        """ Returns True because the underlying data can be sliced.
            The scalar will be wrapped in an array with one element so it can be inspected.
        """
        return True


    def __getitem__(self, index):
        """ Called when using the RTI with an index (e.g. rti[0]).
            The scalar will be wrapped in an array with one element so it can be inspected.
        """
        array = np.array([self._h5Dataset[()]]) # slice with empty tuple
        maskedArray = maskedEqual(array, self.missingDataValue)

        assert maskedArray.shape == (1, ), "Scalar wrapper shape mismatch: {}".format(array.shape)
        return maskedArray[index] # Use the index to ensure the slice has the correct shape


    @property
    def arrayShape(self):
        """ Returns the shape of the wrapper array. Will always be the tuple (1, )
        """
        return (1, )


    @property
    def elementTypeName(self):
        """ String representation of the element type.
        """
        return dataSetElementType(self._h5Dataset)


    @property
    def attributes(self):
        """ The attributes dictionary.
        """
        return self._h5Dataset.attrs


    @property
    def unit(self):
        """ Returns the unit of the RTI by calling dataSetUnit on the underlying dataset
        """
        return dataSetUnit(self._h5Dataset)


    @property
    def missingDataValue(self):
        """ Returns the value to indicate missing data. None if no missing-data value is specified.
        """
        return dataSetMissingValue(self._h5Dataset)



class H5pyFieldRti(BaseRti):
    """ Repository Tree Item (RTI) that contains a field in a structured HDF-5 variable.
    """
    _defaultIconGlyph = RtiIconFactory.FIELD
    _defaultIconColor = ICON_COLOR_H5PY

    def __init__(self, h5Dataset, nodeName, fileName=''):
        """ Constructor.
            The name of the field must be given to the nodeName parameter.
        """
        super(H5pyFieldRti, self).__init__(nodeName, fileName=fileName)
        check_class(h5Dataset, h5py.Dataset)
        self._h5Dataset = h5Dataset


    def hasChildren(self):
        """ Returns False. Field nodes never have children.
        """
        return False


    @property
    def attributes(self):
        """ The attributes dictionary.
            Returns the attributes of the variable that contains this field.
        """
        return self._h5Dataset.attrs


    @property
    def isSliceable(self):
        """ Returns True because the underlying data can be sliced.
        """
        return True


    def __getitem__(self, index):
        """ Called when using the RTI with an index (e.g. rti[0]).
            Applies the index on the NCDF variable that contain this field and then selects the
            current field. In pseudo-code, it returns: self.ncVar[index][self.nodeName].

            If the field itself contains a sub-array it returns:
                self.ncVar[mainArrayIndex][self.nodeName][subArrayIndex]
        """
        mainArrayNumDims = len(self._h5Dataset.shape)
        mainIndex = index[:mainArrayNumDims]
        mainArray = self._h5Dataset.__getitem__(mainIndex)
        fieldArray = mainArray[self.nodeName]
        subIndex = tuple([Ellipsis]) + index[mainArrayNumDims:]
        slicedArray = fieldArray[subIndex]

        return maskedEqual(slicedArray, self.missingDataValue)


    @property
    def nDims(self):
        """ The number of dimensions of the underlying array
        """
        return len(self.arrayShape) # h5py datasets don't have an ndim property


    @property
    def _subArrayShape(self):
        """ Returns the shape of the sub-array
            An empty tuple is returned for regular fields, which have no sub array.
        """
        if self._h5Dataset.dtype.fields is None:
            return tuple() # regular field
        else:
            fieldName = self.nodeName
            fieldDtype = self._h5Dataset.dtype.fields[fieldName][0]
            return fieldDtype.shape


    @property
    def arrayShape(self):
        """ Returns the shape of the underlying array.
            If the field contains a subarray the shape may be longer than 1.
        """
        return self._h5Dataset.shape + self._subArrayShape


    @property
    def elementTypeName(self):
        """ String representation of the element type.
        """
        fieldName = self.nodeName
        return str(self._h5Dataset.dtype.fields[fieldName][0])


    @property
    def dimensionNames(self):
        """ Returns a list with the dimension names of the underlying NCDF variable
        """
        nSubDims = len(self._subArrayShape)
        subArrayDims = ['SubDim{}'.format(dimNr) for dimNr in range(nSubDims)]
        return dimNamesFromDataset(self._h5Dataset) + subArrayDims


    @property
    def unit(self):
        """ Returns the unit of the RTI by calling dataSetUnit on the underlying dataset
        """
        unit = dataSetUnit(self._h5Dataset)
        fieldNames = self._h5Dataset.dtype.names

        # If the missing value attribute is a list with the same length as the number of fields,
        # return the missing value for field that equals the self.nodeName.
        if hasattr(unit, '__len__') and len(unit) == len(fieldNames):
            idx = fieldNames.index(self.nodeName)
            return unit[idx]
        else:
            return unit


    @property
    def missingDataValue(self):
        """ Returns the value to indicate missing data. None if no missing-data value is specified.
        """
        value = dataSetMissingValue(self._h5Dataset)
        fieldNames = self._h5Dataset.dtype.names

        # If the missing value attribute is a list with the same length as the number of fields,
        # return the missing value for field that equals the self.nodeName.
        if hasattr(value, '__len__') and len(value) == len(fieldNames):
            idx = fieldNames.index(self.nodeName)
            return value[idx]
        else:
            return value



class H5pyDatasetRti(BaseRti):
    """ Repository Tree Item (RTI) that contains a HDF5 dataset.

        This includes dimenions scales, which are then displayed with a different icon.
    """
    #_defaultIconGlyph = RtiIconFactory.ARRAY # the iconGlyph property is overridden below
    _defaultIconColor = ICON_COLOR_H5PY

    def __init__(self, h5Dataset, nodeName, fileName=''):
        """ Constructor
        """
        super(H5pyDatasetRti, self).__init__(nodeName, fileName=fileName)
        check_class(h5Dataset, h5py.Dataset)
        self._h5Dataset = h5Dataset
        self._isStructured = bool(self._h5Dataset.dtype.names)


    @property
    def iconGlyph(self):
        """ Shows an Array icon for regular datasets but a dimension icon for dimension scales
        """
        if self._h5Dataset.attrs.get('CLASS', None) == b'DIMENSION_SCALE':
            return RtiIconFactory.DIMENSION
        else:
            return RtiIconFactory.ARRAY


    def hasChildren(self):
        """ Returns True if the variable has a structured type, otherwise returns False.
        """
        return self._isStructured


    @property
    def isSliceable(self):
        """ Returns True because the underlying data can be sliced.
        """
        return True


    def __getitem__(self, index):
        """ Called when using the RTI with an index (e.g. rti[0]).
            Passes the index through to the underlying dataset.
            Converts to a masked array using the missing data value as fill_value
        """
        return maskedEqual(self._h5Dataset.__getitem__(index), self.missingDataValue)


    @property
    def arrayShape(self):
        """ Returns the shape of the underlying array.
        """
        return self._h5Dataset.shape


    @property
    def elementTypeName(self):
        """ String representation of the element type.
        """
        return dataSetElementType(self._h5Dataset)


    @property
    def attributes(self):
        """ The attributes dictionary.
        """
        return self._h5Dataset.attrs


    @property
    def dimensionNames(self):
        """ Returns a list with the dimension names of the underlying HDF-5 dataset.
        """
        return dimNamesFromDataset(self._h5Dataset) # TODO: cache?


    @property
    def unit(self):
        """ Returns the unit of the RTI by calling dataSetUnit on the underlying dataset
        """
        return dataSetUnit(self._h5Dataset)


    @property
    def missingDataValue(self):
        """ Returns the value to indicate missing data. None if no missing-data value is specified.
        """
        return dataSetMissingValue(self._h5Dataset)


    def _fetchAllChildren(self):
        """ Fetches all fields that this variable contains.
            Only variables with a structured data type can have fields.
        """
        assert self.canFetchChildren(), "canFetchChildren must be True"

        childItems = []

        # Add fields
        if self._isStructured:
            for fieldName in self._h5Dataset.dtype.names:
                childItems.append(H5pyFieldRti(self._h5Dataset, nodeName=fieldName,
                                               fileName=self.fileName))

        return childItems


class H5pyGroupRti(BaseRti):
    """ Repository Tree Item (RTI) that contains a HDF-5 group.
    """
    _defaultIconGlyph = RtiIconFactory.FOLDER
    _defaultIconColor = ICON_COLOR_H5PY

    def __init__(self, h5Group, nodeName, fileName=''):
        """ Constructor
        """
        super(H5pyGroupRti, self).__init__(nodeName, fileName=fileName)
        check_class(h5Group, h5py.Group, allow_none=True)

        self._h5Group = h5Group


    @property
    def attributes(self):
        """ The attributes dictionary.
        """
        return self._h5Group.attrs if self._h5Group else {}


    def _fetchAllChildren(self):
        """ Fetches all sub groups and variables that this group contains.
        """
        assert self._h5Group is not None, "dataset undefined (file not opened?)"
        assert self.canFetchChildren(), "canFetchChildren must be True"

        childItems = []

        for childName, h5Child in self._h5Group.items():
            if isinstance(h5Child, h5py.Group):
                childItems.append(H5pyGroupRti(h5Child, nodeName=childName,
                                               fileName=self.fileName))
            elif isinstance(h5Child, h5py.Dataset):
                if len(h5Child.shape) == 0:
                    childItems.append(H5pyScalarRti(h5Child, nodeName=childName,
                                                    fileName=self.fileName))
                else:
                    childItems.append(H5pyDatasetRti(h5Child, nodeName=childName,
                                                     fileName=self.fileName))

            elif isinstance(h5Child, h5py.Datatype):
                #logger.debug("Ignored DataType item: {}".format(childName))
                pass
            else:
                logger.warn("Ignored {}. It has an unexpected HDF-5 type: {}"
                            .format(childName, type(h5Child)))

        return childItems



class H5pyFileRti(H5pyGroupRti):
    """ Reads an HDF-5 file using the h5py package.

        See http://www.h5py.org/
    """
    _defaultIconGlyph = RtiIconFactory.FILE
    _defaultIconColor = ICON_COLOR_H5PY

    def __init__(self, nodeName, fileName=''):
        """ Constructor
        """
        super(H5pyFileRti, self).__init__(None, nodeName, fileName=fileName)
        self._checkFileExists()

    def _openResources(self):
        """ Opens the root Dataset.
        """
        logger.info("Opening: {}".format(self._fileName))
        self._h5Group = h5py.File(self._fileName, 'r')


    def _closeResources(self):
        """ Closes the root Dataset.
        """
        logger.info("Closing: {}".format(self._fileName))
        self._h5Group.close()
        self._h5Group = None



# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Classes for registering plugins.
    This is part of the argos.qt package since it uses QSettings for persistency.
"""

import logging, inspect, os, ast, sys

from argos.info import DEBUGGING
from argos.qt import QtCore
from argos.qt.misc import containsSettingsGroup, removeSettingsGroup
from argos.utils.cls import import_symbol, check_is_a_string, type_name, check_class
from argos.utils.misc import string_to_identifier

logger = logging.getLogger(__name__)


GRP_REGISTRY = 'registry'

def nameToIdentifier(fullName):
    """ Constructs the regItem identifier given its full name
    """
    return string_to_identifier(fullName, white_space_becomes='')

# TODO: why is this in the Qt package? Only be because we use QtCore settings. Refactor?
class ClassRegItem(object):
    """ Represents an class that is registered in the registry.

        Each regItem has a fullName (e.g. Qt/Table) that is used in menus, dialogs, etc. From the
        fullName an identifier is constructed by converting to lower case and removing white space.
        The identifier must be unique. The classFullName is the name of the underlying class,
        including package and module names. If must be accessible from the python-path.

        Each registry item (RegItem) can import its class. If the import fails, the exception info
        is put into the exception property. The underlying class is not imported by default;
        use tryImportClass or getClass() for this.

        Some of the underlying class's attributes, such as the docstring, are made available as
        properties of the RegItem as well. If the class is not yet imported, they return None.
    """
    def __init__(self, fullName, fullClassName, pythonPath=''):
        """ Constructor.

            :param fullName: fullName comprising of library and name, separated by a slash.
                Can contain spaces. E.g.: 'library name/My Widget'
                Must be unique when spaces are removed and converted to lower case.
            :param fullClassName: full name of the underlying class.
                E.g.: 'argos.plugins.rti.ncdf.NcdfFileInspector'
            :param pythonPath: directory that will be added to the sys.path before importing.
                Can be multiple directories separated by a colon (:)
        """
        check_is_a_string(fullClassName)
        self._identifier = nameToIdentifier(fullName)
        self._fullName = fullName
        self._fullClassName = fullClassName
        self._pythonPath = pythonPath

        self._cls = None # The underlying class. Not yet imported.
        self._triedImport = False
        self._exception = None # Any exception that occurs during the class import

    def __repr__(self):
        return "<{} (Ox{:x}): {!r}>".format(type_name(self), id(self), self.fullName)

    @property
    def identifier(self):
        """ Identifier that uniquely identifies the regItem
            Can contain slashes. E.g.: 'libraryname/MyWidget'
        """
        return self._identifier

    @property
    def fullName(self):
        """ Full plugin name comprising of library and name, separated by a slash.
            Can contain spaces. E.g.: 'library name/My Widget'
            Must be unique when spaces are removed and converted to lower case.
        """
        return self._fullName

    @property
    def name(self):
        """ The last part of the fullName.
        """
        return os.path.basename(self._fullName)

    @property
    def library(self):
        """ The fullName minus the last part (the name).
            Used to group libraries together, for instance in menus.
        """
        return os.path.dirname(self._fullName)

    def splitName(self):
        """ Returns (self.library, self.name) tuple but is more efficient than calling both
            properties separately.
        """
        return os.path.split(self._fullName)

    @property
    def fullClassName(self):
        """ full name of the underlying class.
            E.g.: 'argos.plugins.rti.ncdf.NcdfFileInspector'
        """
        return self._fullClassName

#    TODO: not used, remove?
#    @property
#    def className(self):
#        """ The name of the underlying class. Is the last part of the fullClassName.
#        """
#        return self.fullClassName.rsplit('.', 1)[1]

    @property
    def cls(self):
        """ Returns the underlying class.
            Returns None if the class was not imported or import failed.
        """
        return self._cls


    @property
    def pythonPath(self):
        """ Directory that will be added to the sys.path before importing.
            Can be multiple directories separated by a colon (:)
        """
        return self._pythonPath


    @property
    def docString(self):
        """ A cleaned up version of the doc string of the registered class.
            Can serve as backup in case descriptionHtml is empty.
        """
        return inspect.cleandoc('' if self.cls is None else self.cls.__doc__)


    @property
    def descriptionHtml(self):
        """ HTML help describing the class. For use in the detail editor.
        """
        if self.cls is None:
            return None
        elif hasattr(self.cls, 'descriptionHtml'):
            return self.cls.descriptionHtml()
        else:
            return ''


    @property
    def triedImport(self):
        """ Returns True if the class has been imported (either successfully or not)
        """
        return self._triedImport

    @property
    def successfullyImported(self):
        """ Returns True if the import was a success, False if an exception was raised.
            Returns None if the class was not yet imported.
        """
        if self.triedImport:
            return self.exception is None
        else:
            return None

    @property
    def exception(self):
        """ The exception that occurred during the class import.
            Returns None if the import was successful.
        """
        return self._exception


    def tryImportClass(self):
        """ Tries to import the registered class.
            Will set the exception property if and error occurred.
        """
        logger.info("Importing: {}".format(self.fullClassName))
        self._triedImport = True
        self._exception = None
        self._cls = None
        try:
            for pyPath in self.pythonPath.split(':'):
                if pyPath and pyPath not in sys.path:
                    logger.debug("Appending {!r} to the PythonPath.".format(pyPath))
                    sys.path.append(pyPath)
            self._cls = import_symbol(self.fullClassName) # TODO: check class?
        except Exception as ex:
            self._exception = ex
            logger.warn("Unable to import {!r}: {}".format(self.fullClassName, ex))
            if DEBUGGING:
                raise


    def getClass(self, tryImport=True):
        """ Gets the underlying class. Tries to import if tryImport is True (the default).
            Returns None if the import has failed (the exception property will contain the reason)
        """
        if not self.triedImport and tryImport:
            self.tryImportClass()

        return self._cls

    @classmethod
    def createFromDict(cls, dct):
        """ Create an object of type cls with the dct as the kwargs
        """
        return cls(**dct)

    def asDict(self):
        """ Returns a dictionary for serialization. We don't use JSON since all items are
            quite simple and the registry will always contain the same type of ClassRegItem
        """
        return {'fullName': self.fullName,
                'fullClassName': self.fullClassName,
                'pythonPath': self.pythonPath}


# TODO: Each class has an identifier that must be unique in lower-case with spaces are removed?
class ClassRegistry(object):
    """ Class that maintains the collection of registered classes (plugins).

        It can load or store its classes in the persistent QSettings. It can also create a default
        set of plugins that can be used initially, the first time the program is executed.

        The ClassRegistry can only store items of one type (ClassRegItem). Descendants will
        store their own type. For instance the InspectorRegistry will store InspectorRegItem
        items. This makes serialization easier.
    """
    def __init__(self, settingsGroupName=None):
        """ Constructor

            An optional QSettings group name can be specified so that the registry knows where to
            load/store its settings. This can also be specified with the method parameters.
        """
        self.settingsGroupName = settingsGroupName

        # We use an list to store the items in order and an index to find them in O(1)
        # We cannot use an ordereddict for this as this uses linked-list internally and therefore
        # does not allow to retrieve the Nth element in O(1)
        self._items = []
        self._index = {}

        # The registry can only contain items of this type.
        self._itemClass = ClassRegItem


    @property
    def items(self):
        """ The registered class items. Use as read-only
        """
        return self._items


    def clear(self):
        """ Empties the registry
        """
        self._items = []
        self._index = {}


    def getItemById(self, identifier):
        """ Gets a registered item given its identifier. Raises KeyError if not found.
        """
        return self._index[identifier]


    def registerItem(self, regItem):
        """ Adds a ClassRegItem object to the registry.
        """
        check_class(regItem, ClassRegItem)
        if regItem.identifier in self._index:
            oldRegItem = self._index[regItem.identifier]
            logger.warn("Class key {!r} already registered as {}. Removing old regItem."
                        .format(regItem.identifier, oldRegItem.fullClassName))
            self.removeItem(oldRegItem)

        logger.info("Registering {!r} with {}".format(regItem.identifier, regItem.fullClassName))
        self._items.append(regItem)
        self._index[regItem.identifier] = regItem


    def removeItem(self, regItem):
        """ Removes a ClassRegItem object to the registry.
            Will raise a KeyError if the regItem is not registered.
        """
        check_class(regItem, ClassRegItem)
        logger.info("Removing {!r} containing {}".format(regItem.identifier, regItem.fullClassName))

        del self._index[regItem.identifier]
        idx = self._items.index(regItem)
        del self._items[idx]


    def loadOrInitSettings(self, groupName=None):
        """ Reads the registry items from the persistent settings store, falls back on the
            default plugins if there are no settings in the store for this registry.
        """
        groupName = groupName if groupName else self.settingsGroupName
        settings = QtCore.QSettings()

        #for key in sorted(settings.allKeys()):
        #    print(key)

        if containsSettingsGroup(groupName, settings):
            self.loadSettings(groupName)
        else:
            logger.info("Group {!r} not found, falling back on default settings".format(groupName))
            for item in self.getDefaultItems():
                self.registerItem(item)
            self.saveSettings(groupName)
            assert containsSettingsGroup(groupName, settings), \
                "Sanity check failed. {} not found".format(groupName)


    def loadSettings(self, groupName=None):
        """ Reads the registry items from the persistent settings store.
        """
        groupName = groupName if groupName else self.settingsGroupName
        settings = QtCore.QSettings()
        logger.info("Reading {!r} from: {}".format(groupName, settings.fileName()))

        settings.beginGroup(groupName)
        self.clear()
        try:
            for key in settings.childKeys():
                if key.startswith('item'):
                    dct = ast.literal_eval(settings.value(key))
                    regItem = self._itemClass.createFromDict(dct)
                    self.registerItem(regItem)
        finally:
            settings.endGroup()


    def saveSettings(self, groupName=None):
        """ Writes the registry items into the persistent settings store.
        """
        groupName = groupName if groupName else self.settingsGroupName
        settings = QtCore.QSettings()
        logger.info("Saving {} to: {}".format(groupName, settings.fileName()))

        settings.remove(groupName) # start with a clean slate
        settings.beginGroup(groupName)
        try:
            for itemNr, item in enumerate(self.items):
                key = "item-{:03d}".format(itemNr)
                value = repr(item.asDict())
                settings.setValue(key, value)
        finally:
            settings.endGroup()


    def deleteSettings(self, groupName=None):
        """ Deletes registry items from the persistent store.
        """
        groupName = groupName if groupName else self.settingsGroupName
        settings = QtCore.QSettings()
        logger.info("Deleting {} from: {}".format(groupName, settings.fileName()))
        removeSettingsGroup(groupName)


    def getDefaultItems(self):
        """ Returns a list with the default plugins in the registry.
            This is used initialize the application plugins when there are no saved settings,
            for instance the first time the application is started.
            The base implementation returns an empty list but other registries should override it.
        """
        return []





# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Repository Tree Items (RTI) that store Pandas objects.

    See: http://pandas.pydata.org/
"""
import logging
import numpy as np
import pandas as pd

from pandas.core.generic import NDFrame

from argos.repo.baserti import BaseRti
from argos.repo.iconfactory import RtiIconFactory
from argos.utils.cls import check_class

logger = logging.getLogger(__name__)

ICON_COLOR_PANDAS = '#FB9A99'

class PandasIndexRti(BaseRti):
    """ Contains a Pandas undex.
    """
    _defaultIconGlyph = RtiIconFactory.DIMENSION
    _defaultIconColor = RtiIconFactory.COLOR_MEMORY

    def __init__(self, index=None, nodeName='', fileName='',
                 iconColor=_defaultIconColor):
        """ Constructor

            The Index is not part of Pandas' documented API, although it is not private either,
            many metods return a pd.core.index.Index, but the class itself is not documented.
            We therefore don't check if index is of the correct type, as it is not clear if
            pd.core.index.Index is the ancestor of all panda index objects.

            :param index: the underlying pandas index.
        """
        super(PandasIndexRti, self).__init__(nodeName=nodeName, fileName=fileName)

        self._index = index
        self._iconColor = iconColor


    def hasChildren(self):
        """ Returns False. An index never has child nodes.
        """
        return False


    @property
    def isSliceable(self):
        """ Returns True because the underlying index is sliceable.
        """
        return True


    def __getitem__(self, idx):
        """ Called when using the RTI with an index (e.g. rti[0]).
            Returns np.array(index[idx]), where index is the underlying Pandas index.
        """
        return np.array(self._index.__getitem__(idx))


    @property
    def nDims(self):
        """ The number of dimensions of the index. Will always be 1.
        """
        result = self._index.ndim
        assert result == 1, "Expected index to be 1D, got: {}D".format(result)
        return result


    @property
    def arrayShape(self):
        """ Returns the shape of the underlying index.
        """
        return self._index.shape


    @property
    def elementTypeName(self):
        """ String representation of the element type.
        """
        try:
            return str(self._index.dtype) # Series
        except AttributeError:
            assert False, "not yet implemented." # TODO multi-indexes
            return '<structured>' # DataFrames and Panels





class AbstractPandasNDFrameRti(BaseRti):
    """ Contains a Pandas NDFrame object.

        The NDFrame class is the ancestor of pandas Series, DataFrame and Panels.
        May contain None as well (for unopened nodes).

        This is an abstract class and should not be created directly. Use one of its subclasses
        instead.
    """
    # Show always an array, even when _standAlone is false. In this case the NDFrame is a slice of
    # a higher-dimensional NDFrame, so the Field glyph, does not apply (fields have the same nr
    # of dimensions as their parents).
    _defaultIconGlyph = RtiIconFactory.ARRAY
    _defaultIconColor = RtiIconFactory.COLOR_MEMORY

    def __init__(self, ndFrame=None, nodeName='', fileName='', standAlone=True,
                 iconColor=_defaultIconColor):
        """ Constructor

            The NDFrame is not part of Pandas' documented API, although it mentions this
            inheritance. Therefore it is not checked the ndFrame is actually of type NDFrame.

            :param ndFrame: the underlying pandas object. May be undefined (None)
            :type ndFrame: pandas.core.generic.NDFrame
            :param standAlone: True if the NDFrame is a stand-alone object, False if it is part of
                another higher-dimensional, NDFrame. This influences the array. Furthermore, if
                standAlone is True the index of the NDFrame will be included when the children
                are fetched and included in the tree (as a PandasIndexRti)
        """
        super(AbstractPandasNDFrameRti, self).__init__(nodeName=nodeName, fileName=fileName)
        check_class(ndFrame, NDFrame, allow_none=True)

        self._ndFrame = ndFrame
        self._iconColor = iconColor
        self._standAlone = standAlone


    @property
    def _isStructured(self):
        """ Returns True if the class can be subdivided in separate fields.

            Abstract method, will be overridden in the descendants.

            A DataFrame is structured because it is composed of multiple Series. A Series can not
            be subdivided further and thus will return False.
        """
        raise NotImplementedError()


    def hasChildren(self):
        """ Returns True if the variable has a structured type, otherwise returns False.
        """
        return self._ndFrame is not None and (self._isStructured or self._standAlone)


    @property
    def isSliceable(self):
        """ Returns True if the underlying series is not None.
        """
        return self._ndFrame is not None


    def __getitem__(self, index):
        """ Called when using the RTI with an index (e.g. rti[0]).
            Passes the index through to the values property of the underlying NDFrame.
        """
        assert self.isSliceable, "No underlying pandas object: self._ndFrame is None"
        return self._ndFrame.values.__getitem__(index)


    @property
    def nDims(self):
        """ The number of dimensions of the pandas object.
        """
        assert self.isSliceable, "No underlying pandas object: self._ndFrame is None"
        return self._ndFrame.ndim


    @property
    def arrayShape(self):
        """ Returns the shape of the underlying pandas object.
        """
        assert self.isSliceable, "No underlying pandas object: self._ndFrame is None"
        return self._ndFrame.shape


    @property
    def elementTypeName(self):
        """ String representation of the element type.
        """
        if self._ndFrame is None:
            return super(AbstractPandasNDFrameRti, self).elementTypeName
        else:
            try:
                return str(self._ndFrame.dtype) # Series
            except AttributeError:
                return '<structured>' # DataFrames and Panels


    def _createIndexRti(self, index, nodeName):
        """ Auxiliary method that creates a PandasIndexRti.
        """
        return PandasIndexRti(index=index, nodeName=nodeName, fileName=self.fileName,
                              iconColor=self._iconColor)



class PandasSeriesRti(AbstractPandasNDFrameRti):
    """ Contains a Pandas Series.
    """
    @property
    def _isStructured(self):
        """ Returns False, because Series cannot be further subdivided.
        """
        return False


    @property
    def dimensionNames(self):
        """ Returns a list with names that correspond to every NDFrame dimension.
            For Series this is: ['index']
        """
        return ['index']


    def _fetchAllChildren(self):
        """ Fetches the index if the showIndex member is True

            Descendants can override this function to add the subdevicions.
        """
        assert self.isSliceable, "No underlying pandas object: self._ndFrame is None"
        childItems = []
        if self._standAlone:
            childItems.append(self._createIndexRti(self._ndFrame.index, 'index'))
        return childItems



class PandasDataFrameRti(AbstractPandasNDFrameRti):
    """ Contains a Pandas DataFrame
    """
    @property
    def _isStructured(self):
        """ Returns True, because DataFrames consist of Series
        """
        return True


    @property
    def dimensionNames(self):
        """ Returns a list with names that correspond to every NDFrame dimension.
            For DataFrames this is: ['index', 'columns']
        """
        return ['index', 'columns']


    def _fetchAllChildren(self):
        """ Fetches children items.

            If this is stand-alone DataFrame the index, column etc are added as PandasIndexRti obj.
        """
        assert self.isSliceable, "No underlying pandas object: self._ndFrame is None"

        childItems = []

        for subName in self._ndFrame.columns: # Note that this is not the first dimension!
            childItem = PandasSeriesRti(self._ndFrame[subName], nodeName=subName,
                                        fileName=self.fileName, iconColor=self._iconColor,
                                        standAlone=False)
            childItems.append(childItem)

        if self._standAlone:
            childItems.append(self._createIndexRti(self._ndFrame.index, 'index'))
            childItems.append(self._createIndexRti(self._ndFrame.columns, 'columns'))

        return childItems



class PandasPanelRti(AbstractPandasNDFrameRti):
    """ Contains a Pandas Panel
    """
    @property
    def _isStructured(self):
        """ Returns True, because Panels consist of DataFrames
        """
        return True


    @property
    def dimensionNames(self):
        """ Returns a list with names that correspond to every NDFrame dimension.
            For DataFrames this is: ['items', 'major_axis', 'minor_axis'], which correspond to the
            data frames itself, the rows and the columns of the underlying DataFrames.
        """
        return ['items', 'major_axis', 'minor_axis']


    def _fetchAllChildren(self):
        """ Fetches children items.

            If this is stand-alone Panel the index, column etc are added as PandasIndexRti obj.
        """
        assert self.isSliceable, "No underlying pandas object: self._ndFrame is None"

        childItems = []

        for subName in self._ndFrame.items:
            childItem = PandasDataFrameRti(self._ndFrame[subName], nodeName=subName,
                                           fileName=self.fileName, iconColor=self._iconColor,
                                           standAlone=False)
            childItems.append(childItem)

        if self._standAlone:
            childItems.append(self._createIndexRti(self._ndFrame.items, 'items'))
            childItems.append(self._createIndexRti(self._ndFrame.major_axis, 'major_axis'))
            childItems.append(self._createIndexRti(self._ndFrame.minor_axis, 'minor_axis'))

        return childItems



class PandasCsvFileRti(PandasDataFrameRti):
    """ Reads a comma-separated file (CVS) into a Pandas DataFrame.
    """
    _defaultIconGlyph = RtiIconFactory.FILE
    _defaultIconColor = ICON_COLOR_PANDAS

    def __init__(self, nodeName='', fileName=''):
        """ Constructor. Initializes as an ArrayRTI with None as underlying array.
        """
        super(PandasCsvFileRti, self).__init__(ndFrame=None, nodeName=nodeName, fileName=fileName,
                                               iconColor=PandasCsvFileRti._defaultIconColor,
                                               standAlone=True)
        self._checkFileExists()


    def hasChildren(self):
        """ Returns True so that a triangle is added that expands the node and opens the file
        """
        return True


    def _openResources(self):
        """ Uses numpy.loadtxt to open the underlying file
        """
        self._ndFrame = pd.read_csv(self._fileName)


    def _closeResources(self):
        """ Closes the underlying resources
        """
        self._ndFrame = None




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Version and other info for this program
"""
import logging
import sys

from argos.info import DEBUGGING
from argos.inspector.registry import InspectorRegistry, DEFAULT_INSPECTOR
from argos.qt import QtCore, QtWidgets, QtSlot
from argos.qt.misc import removeSettingsGroup, handleException, initQApplication
from argos.qt.registry import GRP_REGISTRY, nameToIdentifier
from argos.repo.registry import globalRtiRegistry
from argos.repo.repotreemodel import RepoTreeModel
from argos.utils.misc import string_to_identifier
from argos.widgets.mainwindow import MainWindow, UpdateReason

logger = logging.getLogger(__name__)


class ArgosApplication(QtCore.QObject):
    """ The application singleton which holds global state.
    """
    def __init__(self, setExceptHook=True):
        """ Constructor

            :param setExceptHook: Sets the global sys.except hook so that Qt shows a dialog box
                when an exception is raised.

                In debugging mode, the program will just quit in case of an exception. This is
                standard Python behavior but PyQt and PySide swallow exceptions by default (only a
                log message is displayed). The practice of swallowing exceptions fosters bad
                programming IHMO as it is easy to miss errors. I strongly recommend that you set
                the setExceptHook to True.
        """
        super(ArgosApplication, self).__init__()

        # Call initQtWidgetsApplicationInstance() so that the users can call argos.browse without
        # having to call it themselves.
        self._qApplication = initQApplication()

        if setExceptHook:
            logger.debug("Setting sys.excepthook to Argos exception handling")
            sys.excepthook = handleException

        #self.qApplication.focusChanged.connect(self.focusChanged) # for debugging

        self._repo = RepoTreeModel()
        self._rtiRegistry = globalRtiRegistry()
        self._inspectorRegistry = InspectorRegistry()

        self._profile = ''
        self._mainWindows = []
        self._settingsSaved = False  # boolean to prevent saving settings twice

        self.qApplication.lastWindowClosed.connect(self.quit)

        # Activate-actions for all windows
        self.windowActionGroup = QtWidgets.QActionGroup(self)
        self.windowActionGroup.setExclusive(True)

        # Call setup when the event loop starts.
        QtCore.QTimer.singleShot(0, self.setup)


    def setup(self):
        """ Called once directly after the event loop starts.
        """
        logger.debug("ArgosApplication.setup called")

        # Raising all window because in OS-X window 0 is not shown.
        #self.raiseAllWindows()
        # activateWindow also solves the issue but doesn't work with the --inspector option.
        self.windowActionGroup.actions()[0].trigger()


    ##############
    # Properties #
    ##############

    @property
    def qApplication(self):
        """ Returns the QApplication object. Equivalent to QtWidgets.qApp.
        """
        return self._qApplication


    @property
    def repo(self):
        """ Returns the global repository
        """
        return self._repo


    @property
    def rtiRegistry(self):
        """ Returns the repository tree item (rti) registry
        """
        return self._rtiRegistry


    @property
    def inspectorRegistry(self):
        """ Returns the repository tree item (rti) registry
        """
        return self._inspectorRegistry


    @property
    def profile(self):
        """ Persistent settings are associated to a profile. This allows users to save the
            program state for several usage profiles.
            Profile settings are case insensitive.
        """
        return self._profile


    @property
    def mainWindows(self):
        """ Returns the list of MainWindows. For read-only purposes only.
        """
        return self._mainWindows

    ###########
    # Methods #
    ###########

    def focusChanged(self, old, now):
        """ Is called when the focus changes. Useful for debugging.
        """
        logger.debug("Focus changed from {} to {}".format(old, now))


    def deleteRegistries(self):
        """ Deletes all registry information from the persistent store.
        """
        removeSettingsGroup(GRP_REGISTRY)


    def loadOrInitRegistries(self):
        """ Reads the registry persistent program settings
        """
        self.inspectorRegistry.loadOrInitSettings()
        self.rtiRegistry.loadOrInitSettings()


#    def saveRegistries(self):
#        """ Writes the view settings to the persistent store
#        """
#        self.rtiRegistry.saveSettings(self.GRP_REGISTRY_RTI)
#        self.inspectorRegistry.saveSettings(self.GRP_REGISTRY_INSPECTORS)


    def profileGroupName(self, profile=None):
        """ Returns the name of the QSetting group for the profile.
            Converts to lower case and removes whitespace, interpunction, etc.
            Prepends _debug_ if the debugging flag is set

            :param profile: profile name. If None the current profile is used.
        """
        profile = profile if profile else self.profile
        profGroupName = '_debug_' if DEBUGGING else ''
        profGroupName += string_to_identifier(profile)
        return profGroupName


    def windowGroupName(self, windowNumber, profile=None):
        """ Returns the name of the QSetting group for this window in the this profile.

            :param windowNumber: int
            :param profile: profile name. If None the current profile is used.
        """
        return "{}/window-{:02d}".format(self.profileGroupName(profile=profile), windowNumber)


    def deleteProfile(self, profile):
        """ Removes a profile from the persistent settings
        """
        profGroupName = self.profileGroupName(profile)
        logger.debug("Resetting profile settings: {}".format(profGroupName))
        settings = QtCore.QSettings()
        settings.remove(profGroupName)


    def deleteAllProfiles(self):
        """ Returns a list of all profiles
        """
        settings = QtCore.QSettings()
        for profGroupName in QtCore.QSettings().childGroups():
            settings.remove(profGroupName)


    def loadProfile(self, profile, inspectorFullName=None):
        """ Reads the persistent program settings for the current profile.

            If inspectorFullName is given, a window with this inspector will be created if it wasn't
            already created in the profile. All windows with this inspector will be raised.
        """
        settings = QtCore.QSettings()
        logger.info("Reading profile {!r} from: {}".format(profile, settings.fileName()))

        self._profile = profile
        profGroupName = self.profileGroupName(profile)

        # Instantiate windows from groups
        settings.beginGroup(profGroupName)
        try:
            for windowGroupName in settings.childGroups():
                if windowGroupName.startswith('window'):
                    settings.beginGroup(windowGroupName)
                    try:
                        self.addNewMainWindow(settings=settings)
                    finally:
                        settings.endGroup()
        finally:
            settings.endGroup()

        if inspectorFullName is not None:
            windows = [win for win in self._mainWindows
                       if win.inspectorFullName == inspectorFullName]
            if len(windows) == 0:
                logger.info("Creating window for inspector: {!r}".format(inspectorFullName))
                try:
                    win = self.addNewMainWindow(inspectorFullName=inspectorFullName)
                except KeyError:
                    logger.warn("No inspector found with ID: {}".format(inspectorFullName))
            else:
                for win in windows:
                    win.raise_()

        if len(self.mainWindows) == 0:
            logger.info("No open windows in profile (creating one).")
            self.addNewMainWindow(inspectorFullName=DEFAULT_INSPECTOR)



    def saveProfile(self):
        """ Writes the current profile settings to the persistent store
        """
        if not self.profile:
            logger.warning("No profile defined (no settings saved)")
            return

        settings = QtCore.QSettings()
        logger.debug("Writing settings to: {}".format(settings.fileName()))

        profGroupName = self.profileGroupName()
        settings.remove(profGroupName) # start with a clean slate

        assert self.mainWindows, "no main windows found"
        for winNr, mainWindow in enumerate(self.mainWindows):
            settings.beginGroup(self.windowGroupName(winNr))
            try:
                mainWindow.saveProfile(settings)
            finally:
                settings.endGroup()


    def saveSettings(self):
        """ Saves the persistent settings. Only saves the profile.
        """
        try:
            self.saveProfile()
        except Exception as ex:
            # Continue, even if saving the settings fails.
            logger.warn(ex)
            if DEBUGGING:
                raise
        finally:
            self._settingsSaved = True


    def saveSettingsIfNeeded(self):
        """ Writes the persistent settings of this profile is this is the last window and
            the settings have not yet been saved.
        """
        if not self._settingsSaved and len(self.mainWindows) <= 1:
            self.saveSettings()


    def loadFiles(self, fileNames, rtiClass=None):
        """ Loads files into the repository as repo tree items of class rtiClass.
            Auto-detects using the extensions when rtiClass is None
        """
        for fileName in fileNames:
            self.repo.loadFile(fileName, rtiClass=rtiClass)


    def repopulateAllWindowMenus(self):
        """ Repopulates the Window menu of all main windows from scratch.

            To be called when a main window is created or removed.
        """
        for win in self.mainWindows:
            win.repopulateWinowMenu(self.windowActionGroup)


    @QtSlot()
    def addNewMainWindow(self, settings=None, inspectorFullName=None):
        """ Creates and shows a new MainWindow.

            If inspectorFullName is set, it will set the identifier from that name.
            If the inspector identifier is not found in the registry, a KeyError is raised.
        """
        mainWindow = MainWindow(self)
        self.mainWindows.append(mainWindow)

        self.windowActionGroup.addAction(mainWindow.activateWindowAction)
        self.repopulateAllWindowMenus()

        if settings:
            mainWindow.readViewSettings(settings)

        if inspectorFullName:
            inspectorId = nameToIdentifier(inspectorFullName)
            mainWindow.setInspectorById(inspectorId)

        if mainWindow.inspectorRegItem: # can be None at start
            inspectorId = mainWindow.inspectorRegItem.identifier
            mainWindow.getInspectorActionById(inspectorId).setChecked(True)
            logger.info("Created new window with inspector: {}"
                        .format(mainWindow.inspectorRegItem.fullName))
        else:
            logger.info("Created new window without inspector")

        mainWindow.drawInspectorContents(reason=UpdateReason.NEW_MAIN_WINDOW)
        mainWindow.show()

        if sys.platform.startswith('darwin'):
            # Calling raise before the QApplication.exec_ only shows the last window
            # that was added. Therefore we also call activeWindow. However, this may not
            # always be desirable. TODO: make optional?
            mainWindow.raise_()
            pass

        return mainWindow


    def removeMainWindow(self, mainWindow):
        """ Removes the mainWindow from the list of windows. Saves the settings
        """
        logger.debug("removeMainWindow called")

        self.windowActionGroup.removeAction(mainWindow.activateWindowAction)
        self.repopulateAllWindowMenus()

        self.mainWindows.remove(mainWindow)


    def raiseAllWindows(self):
        """ Raises all application windows.
        """
        logger.debug("raiseAllWindows called")
        for mainWindow in self.mainWindows:
            logger.debug("Raising {}".format(mainWindow._instanceNr))
            mainWindow.raise_()


    def closeAllWindows(self):
        """ Closes all windows. Save windows state to persistent settings before closing them.
        """
        self.saveSettings()
        logger.debug("ArgosApplication: Closing all windows")
        self.qApplication.closeAllWindows()


    def quit(self):
        """ Quits the application (called when the last window is closed)
        """
        logger.debug("ArgosApplication.quit called")
        assert len(self.mainWindows) == 0, \
            "Bug: still {} windows present at application quit!".format(len(self.mainWindows))
        self.qApplication.quit()


    def execute(self):
        """ Executes all main windows by starting the Qt main application
        """
        logger.info("Starting Argos event loop...")
        exitCode = self.qApplication.exec_()
        logger.info("Argos event loop finished with exit code: {}".format(exitCode))
        return exitCode




# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Common functionality, look and feel for all tree views in Argos.

    Currently it supports only selecting one item. That is, the current item is always the
    selected item.

    The difference between the current item and the selected item(s) is as follows:
    (from: http://doc.qt.digia.com/4.6/model-view-selection.html)

    In a view, there is always a current item and a selected item - two independent states.
    An item can be the current item and selected at the same time. The view is responsible for
    ensuring that there is always a current item as keyboard navigation, for example, requires
    a current item.

    Current Item:
        There can only be one current item.
        The current item will be changed with key navigation or mouse button clicks.
        The current item will be edited if the edit key, F2, is pressed or the item is
            double-clicked (provided that editing is enabled).
        The current item is indicated by the focus rectangle.

    Selected Items:
        There can be multiple selected items.
        The selected state of items is set or unset, depending on several pre-defined modes
            (e.g., single selection, multiple selection, etc.) when the user interacts with the
            items.
        The current item can be used together with an anchor to specify a range that should be
            selected or deselected (or a combination of the two).
        The selected items are indicated with the selection rectangle.
"""
from __future__ import print_function

import logging
from argos.qt import QtCore, QtWidgets
from argos.qt.togglecolumn import ToggleColumnTreeView
from argos.qt.treemodels import BaseTreeModel
from argos.utils.cls import check_class
from argos.widgets.constants import TREE_ICON_SIZE


logger = logging.getLogger(__name__)

# Qt classes have many ancestors
#pylint: disable=R0901

class ArgosTreeView(ToggleColumnTreeView):
    """ QTreeView that defines common functionality, look and feel for all tree views in Argos.

        The model must be a BaseTreeModel
    """
    def __init__(self, treeModel=None, parent=None):
        """ Constructor
        """
        super(ArgosTreeView, self).__init__(parent)

        if treeModel is not None:
            self.setModel(treeModel)

        self.setAlternatingRowColors(True)
        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.setAnimated(True)
        self.setAllColumnsShowFocus(True)
        self.setIconSize(TREE_ICON_SIZE)

        treeHeader = self.header()
        treeHeader.setSectionsMovable(True)
        treeHeader.setSectionResizeMode(QtWidgets.QHeaderView.Interactive) # don't set to stretch
        treeHeader.setStretchLastSection(True)


    def setModel(self, model):
        """ Sets the model.
            Checks that the model is a
        """
        check_class(model, BaseTreeModel)
        super(ArgosTreeView, self).setModel(model)


    def setCurrentIndex(self, currentIndex):
        """ Sets the current item to be the item at currentIndex.
            Also select the row as to give consistent user feedback.
            See also the notes at the top of this module on current item vs selected item(s).
        """
        selectionModel = self.selectionModel()
        selectionFlags = (QtCore.QItemSelectionModel.ClearAndSelect |
                          QtCore.QItemSelectionModel.Rows)
        selectionModel.setCurrentIndex(currentIndex, selectionFlags)


    def getRowCurrentIndex(self):
        """ Returns the index of column 0 of the current item in the underlying model.
            See also the notes at the top of this module on current item vs selected item(s).
        """
        curIndex = self.currentIndex()
        col0Index = curIndex.sibling(curIndex.row(), 0)
        return col0Index


    def getCurrentItem(self): # TODO: rename? getCurrentItemAndIndex? getCurrentTuple? getCurrent?
        """ Find the current tree item (and the current index while we're at it)
            Returns a tuple with the current item, and its index. The item may be None.
            See also the notes at the top of this module on current item vs selected item(s).
        """
        currentIndex = self.getRowCurrentIndex()
        currentItem = self.model().getItem(currentIndex)
        return currentItem, currentIndex


    def expandPath(self, path):
        """ Follows the path and expand all nodes along the way.
            Returns (item, index) tuple of the last node in the path (the leaf node). This can be
            reused e.g. to select it.
        """
        iiPath = self.model().findItemAndIndexPath(path)
        for (item, index) in iiPath[1:]: # skip invisible root
            assert index.isValid(), "Sanity check: invalid index in path for item: {}".format(item)
            self.expand(index)

        leaf = iiPath[-1]
        return leaf


    def expandBranch(self, index=None, expanded=True):
        """ Expands or collapses the node at the index and all it's descendants.

            If expanded is True the nodes will be expanded, if False they will be collapsed.

            If parentIndex is None, the invisible root will be used (i.e. the complete forest will
            be expanded).
        """
        treeModel = self.model()
        if index is None:
            index = QtCore.QModelIndex()

        if index.isValid():
            self.setExpanded(index, expanded)

        for rowNr in range(treeModel.rowCount(index)):
            childIndex = treeModel.index(rowNr, 0, parentIndex=index)
            self.expandBranch(index=childIndex, expanded=expanded)




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Version and other info for this program
"""
import os

DEBUGGING = False

VERSION = '0.2.1rc1'
REPO_NAME = "argos"
#SCRIPT_NAME = "argos"
PACKAGE_NAME = "argos"
PROJECT_NAME = "Argos"
DEFAULT_PROFILE = 'Panoptes'
SHORT_DESCRIPTION = "Argos Panoptes HDF/NCDF/scientific data viewer."
PROJECT_URL = "https://github.com/titusjan/argos"
AUTHOR = "Pepijn Kenter"
EMAIL = "titusjan@gmail.com"
ORGANIZATION_NAME = "titusjan"
ORGANIZATION_DOMAIN = "titusjan.nl"


def program_directory():
    """ Returns the program directory where this program is installed
    """
    return os.path.abspath(os.path.dirname(__file__))

def icons_directory():
    """ Returns the program directory where this program is installed
    """
    return os.path.join(program_directory(), 'img/snipicons')




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Miscellaneous routines.
"""
import logging, sys
logger = logging.getLogger(__name__)


class NotSpecified(object):
    """ Class for NOT_SPECIFIED constant.
        Is used so that a parameter can have a default value other than None.

        Evaluate to False when converted to boolean.
    """
    def __nonzero__(self):
        """ Always returns False. Called when to converting to bool in Python 2.
        """
        return False

    def __bool__(self):
        """ Always returns False. Called when to converting to bool in Python 3.
        """
        return False



NOT_SPECIFIED = NotSpecified()



def python_major_version():
    """ Returns 2 or 3 for Python 2.x or 3.x respectively
    """
    return sys.version_info[0]


def python2():
    """ Returns True if we are running python 2
    """
    major_version = sys.version_info[0]
    assert major_version == 2 or major_version == 3, "major_version = {!r}".format(major_version)
    return major_version == 2


def log_dictionary(dictionary, msg='', logger=None, level='debug', item_prefix='  '):
    """ Writes a log message with key and value for each item in the dictionary.

        :param dictionary: the dictionary to be logged
        :type dictionary: dict
        :param name: An optional message that is logged before the contents
        :type name: string
        :param logger: A logging.Logger object to log to. If not set, the 'main' logger is used.
        :type logger: logging.Logger or a string
        :param level: log level. String or int as described in the logging module documentation.
            Default: 'debug'.
        :type level: string or int
        :param item_prefix: String that will be prefixed to each line. Default: two spaces.
        :type item_prefix: string
    """
    level_nr = logging.getLevelName(level.upper())

    if logger is None:
        logger = logging.getLogger('main')

    if msg :
        logger.log(level_nr, "Logging dictionary: {}".format(msg))

    if not dictionary:
        logger.log(level_nr,"{}<empty dictionary>".format(item_prefix))
        return

    max_key_len = max([len(k) for k in dictionary.keys()])

    for key, value in sorted(dictionary.items()):
        logger.log(level_nr, "{0}{1:<{2}s} = {3}".format(item_prefix, key, max_key_len, value))



def prepend_point_to_extension(extension):
    """ Prepends a point to the extension of it doesn't already start with it
    """
    if extension.startswith('.'):
        return extension
    else:
        return '.' + extension


def is_quoted(s):
    """ Returns True if the string begins and ends with quotes (single or double)

        :param s: a string
        :return: boolean
    """
    return (s.startswith("'") and s.endswith("'")) or (s.startswith('"') and s.endswith('"'))



def string_to_identifier(s, white_space_becomes='_'):
    """ Takes a string and makes it suitable for use as an identifier

        Translates to lower case
        Replaces white space by the white_space_becomes character (default=underscore).
        Removes and punctuation.
    """
    import re
    s = s.lower()
    s = re.sub(r"\s+", white_space_becomes, s) # replace whitespace with underscores
    s = re.sub(r"-", "_", s) # replace hyphens with underscores
    s = re.sub(r"[^A-Za-z0-9_]", "", s) # remove everything that's not a character, a digit or a _
    return s


if __name__ == "__main__":
    print (string_to_identifier("Pea\nsdf-43q45,.!@#%&@&@@24n  pijn  Kenter, hallo$"))






# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Contains the StringCti and StringCtiEditor classes
"""
import logging

from argos.config.abstractcti import AbstractCti, AbstractCtiEditor
from argos.qt import QtWidgets

logger = logging.getLogger(__name__)


class StringCti(AbstractCti):
    """ Config Tree Item to store a string. It can be edited with a QLineEdit.
        The string can have an optional maximum length.
    """
    def __init__(self, nodeName, defaultData='', maxLength=None):
        """ For the (other) parameters see the AbstractCti constructor documentation.

            :param maxLength: maximum length of the string
        """
        super(StringCti, self).__init__(nodeName, defaultData)

        # We could define a mask here as well but since that very likely will be rarely used,
        # we don't want to store it for each cti. You can make a subclass if you need it.
        self.maxLength = maxLength

    def _enforceDataType(self, data):
        """ Converts to str so that this CTI always stores that type.
        """
        return str(data)

    @property
    def debugInfo(self):
        """ Returns the string with debugging information
        """
        return "maxLength = {}".format(self.maxLength)

    def createEditor(self, delegate, parent, option):
        """ Creates a StringCtiEditor.
            For the parameters see the AbstractCti constructor documentation.
        """
        return StringCtiEditor(self, delegate, parent=parent)



class StringCtiEditor(AbstractCtiEditor):
    """ A CtiEditor which contains a QLineEdit for editing StringCti objects.
    """
    def __init__(self, cti, delegate, parent=None):
        """ See the AbstractCtiEditor for more info on the parameters
        """
        super(StringCtiEditor, self).__init__(cti, delegate, parent=parent)
        self.lineEditor = self.addSubEditor(QtWidgets.QLineEdit(), isFocusProxy=True)

        if cti.maxLength is not None:
            self.lineEditor.setMaxLength(cti.maxLength)


    def setData(self, data):
        """ Provides the main editor widget with a data to manipulate.
        """
        self.lineEditor.setText(str(data))


    def getData(self):
        """ Gets data from the editor widget.
        """
        return self.lineEditor.text()




# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Collector Tree.
"""
from __future__ import print_function

import logging

from argos.qt import Qt, QtCore, QtGui, QtWidgets, QtSlot
from argos.qt.togglecolumn import ToggleColumnTreeView
from argos.widgets.constants import COLLECTOR_TREE_ICON_SIZE

logger = logging.getLogger(__name__)

# Qt classes have many ancestors
#pylint: disable=R0901

#TODO: do we still need this as a separate class?
class CollectorTree(ToggleColumnTreeView):
    """ Tree widget for collecting the selected data. Includes an internal tree model.

        NOTE: this class is not meant to be used directly but is 'private' to the Collector().
        That is, plugins should interact with the Collector class, not the CollectorTree()
    """
    HEADERS = ["item path", "item name"] # TODO: this can't be right. Is this even used?
    (COL_ITEM_PATH, COL_ITEM_NAME) = range(len(HEADERS))


    def __init__(self, parent):
        """ Constructor
        """
        super(CollectorTree, self).__init__(parent=parent)

        self._comboLabels = None

        nCols = 1
        model = QtGui.QStandardItemModel(3, nCols)
        self.setModel(model)
        self.setTextElideMode(Qt.ElideMiddle) # ellipsis appear in the middle of the text

        self.setRootIsDecorated(False) # disable expand/collapse triangle
        self.setUniformRowHeights(True)

        self.setAlternatingRowColors(True)
        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.setAnimated(True)
        self.setAllColumnsShowFocus(True)
        self.setContextMenuPolicy(Qt.ActionsContextMenu)
        self.setIconSize(COLLECTOR_TREE_ICON_SIZE)

        treeHeader = self.header()
        treeHeader.setStretchLastSection(False)
        treeHeader.setSectionsMovable(False)

        treeHeader.resizeSection(0, 400) # For item path
        treeHeader.setSectionResizeMode(QtWidgets.QHeaderView.Interactive) # don't set to stretch

        labels = [''] * model.columnCount()
        labels[0] = "path"
        model.setHorizontalHeaderLabels(labels)

        #enabled = dict((name, False) for name in self.HEADERS)
        #checked = dict((name, True) for name in self.HEADERS)
        #self.addHeaderContextMenu(checked=checked, enabled=enabled, checkable={})


    def resizeColumnsToContents(self, startCol=None, stopCol=None):
        """ Resizes all columns to the contents
        """
        numCols = self.model().columnCount()
        startCol = 0 if startCol is None else max(startCol, 0)
        stopCol  = numCols if stopCol is None else min(stopCol, numCols)

        row = 0
        for col in range(startCol, stopCol):
            indexWidget = self.indexWidget(self.model().index(row, col))

            if indexWidget:
                contentsWidth = indexWidget.sizeHint().width()
            else:
                contentsWidth = self.header().sectionSizeHint(col)

            self.header().resizeSection(col, contentsWidth)



class CollectorSpinBox(QtWidgets.QSpinBox):
    """ A spinbox for use in the collector.
        Overrides the sizeHint so that is not truncated when large dimension names are selected
    """
    def __init__(self, *args, **kwargs):
        """ Constructor
        """
        super(CollectorSpinBox, self).__init__(*args, **kwargs)
        self._cachedSizeHint = None


    def sizeHint(self):
        """ Reimplemented from the C++ Qt source of QAbstractSpinBox.sizeHint, but without
            truncating to a maximum of 18 characters.
        """
        # The cache is invalid after the prefix, postfix and other properties
        # have been set. I disabled it because sizeHint isn't called that often.
        #if self._cachedSizeHint is not None:
        #    return self._cachedSizeHint

        orgSizeHint = super(CollectorSpinBox, self).sizeHint()

        self.ensurePolished()
        d = self
        fm = QtGui.QFontMetrics(self.fontMetrics())

        # This was h = d.edit.sizeHint().height(), but that didn't work. In the end we set the
        # height to the height calculated from the parent.
        h = orgSizeHint.height()
        w = 0

        # QLatin1Char seems not to be implemented.
        # Using regular string literals and hope for the best
        s = d.prefix() + d.textFromValue(d.minimum()) + d.suffix() + ' '

        # We disabled truncating the string here!!
        #s = s[:18]
        w = max(w, fm.width(s))

        s = d.prefix() + d.textFromValue(d.maximum()) + d.suffix() + ' '

        # We disabled truncating the string here!!
        #s = s[:18]
        w = max(w, fm.width(s))
        if len(d.specialValueText()):
            s = d.specialValueText()
            w = max(w, fm.width(s))

        w += 2 # cursor blinking space

        opt = QtWidgets.QStyleOptionSpinBox()
        self.initStyleOption(opt)
        hint = QtCore.QSize(w, h)
        extra = QtCore.QSize(35, 6)

        opt.rect.setSize(hint + extra)
        extra += hint - self.style().subControlRect(QtWidgets.QStyle.CC_SpinBox, opt,
                                                    QtWidgets.QStyle.SC_SpinBoxEditField, self).size()

        # get closer to final result by repeating the calculation
        opt.rect.setSize(hint + extra)
        extra += hint - self.style().subControlRect(QtWidgets.QStyle.CC_SpinBox, opt,
                                                    QtWidgets.QStyle.SC_SpinBoxEditField, self).size()
        hint += extra

        opt.rect = self.rect()
        result = (self.style().sizeFromContents(QtWidgets.QStyle.CT_SpinBox, opt, hint, self)
                  .expandedTo(QtWidgets.QApplication.globalStrut()))
        self._cachedSizeHint = result

        # Use the height ancestor's sizeHint
        result.setHeight(orgSizeHint.height())

        return result





# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Contains the Config and GroupCtiEditor classes
"""
import logging

from argos.config.abstractcti import AbstractCti, AbstractCtiEditor
from argos.qt import Qt, QtGui, QtWidgets

logger = logging.getLogger(__name__)




class GroupCti(AbstractCti):
    """ Read only config Tree Item that only stores None. It can be used to group CTIs
    """
    def __init__(self, nodeName, defaultData=None, expanded=True):
        """ Constructor. For the parameters see the AbstractCti constructor documentation.
        """
        super(GroupCti, self).__init__(nodeName, defaultData, expanded=expanded)


    def _enforceDataType(self, data):
        """ Passes the data as is; no conversion.
        """
        return data


    def _dataToString(self, data):
        """ Conversion function used to convert the (default)data to the display value.
            Returns an empty string.
        """
        return ""


    def createEditor(self, delegate, parent, _option):
        """ Creates a hidden widget so that only the reset button is visible during editing.
            :type option: QStyleOptionViewItem
        """
        return GroupCtiEditor(self, delegate, parent=parent)


class GroupCtiEditor(AbstractCtiEditor):
    """ A CtiEditor which contains a hidden widget.
        If the item is editable, the reset button is shown.
    """
    def __init__(self, cti, delegate, parent=None):
        """ See the AbstractCtiEditor for more info on the parameters
        """
        super(GroupCtiEditor, self).__init__(cti, delegate, parent=parent)

        # Add hidden widget to store editor value
        self.widget = self.addSubEditor(QtWidgets.QWidget())
        self.widget.hide()


    def setData(self, data):
        """ Provides the main editor widget with a data to manipulate.
        """
        # Set the data in the 'editor_data' property of the widget to that getData
        # can pass the same value back into the CTI.
        self.widget.setProperty("editor_data", data)


    def getData(self):
        """ Gets data from the editor widget.
        """
        return self.widget.property("editor_data")



class MainGroupCti(GroupCti):
    """ Read only config Tree Item that only stores None.
        To be used as a high level group (e.g. the inspector group)
        Is the same as a groupCti but drawn as light text on a dark grey back ground
    """
    _backgroundBrush = QtGui.QBrush(QtGui.QColor("#606060")) # create only once
    _foregroundBrush = QtGui.QBrush(QtGui.QColor(Qt.white)) # create only once
    _font = QtGui.QFont()
    _font.setWeight(QtGui.QFont.Bold)

    def __init__(self, nodeName, defaultData=None):
        """ Constructor. For the parameters see the AbstractCti constructor documentation.
        """
        super(MainGroupCti, self).__init__(nodeName, defaultData, expanded=True) # always expand


    @property
    def font(self):
        """ Returns a font for displaying this item's text in the tree.
        """
        return self._font


    @property
    def backgroundBrush(self):
        """ Returns a (dark gray) brush for drawing the background role in the tree.
        """
        return self._backgroundBrush


    @property
    def foregroundBrush(self):
        """ Returns a (white) brush for drawing the foreground role in the tree.
        """
        return self._foregroundBrush



